%{

#include    <stdio.h>
#include    <string.h>
#include    <ctype.h>

#include    "defs.h"
#include    "lib.h"
#include    "token.h"
#include    "vcpu.h"
#include    "code.h"

%}

%token	BREAK	CASE	CHAR	CONTINUE DEFAULT
%token	DEFINE	DO	DOUBLE	ELSE	EXTERN
%token	FLOAT	FOR	GOTO	IF	IFDEF
%token	IFNDEF	INCLUDE	INT	LONG	REGISTER
%token	RETURN	SHORT	SIGNED	SIZEOF	STATIC
%token	STRUCT	SWITCH	TYPEDEF	UNDEF	UNION
%token	UNSIGNED VOID	WHILE	ANDAND	OROR
%token	INC	DEC	ADDLET	SUBLET	MULLET
%token	DIVLET	MODLET	ANDLET	ORLET	XORLET
%token	EQU	UNEQU	LEFTLET	RIGHTLET LEFT
%token	RIGHT	SML	BIG	EXPR	LABEL
%token	MENBA	STRING	DEXPR
%token	LINC	LDEC

%start	program

%nonassoc IF
%nonassoc ELSE

%left	LOW
%left	','
%right	'=' ANDLET ORLET XORLET ADDLET SUBLET MULLET DIVLET MODLET LEFTLET RIGHTLET
%right	'?'
%left	OROR
%left	ANDAND
%left	'|'
%left	'&'
%left	'^'
%left	EQU	UNEQU
%left	'<'	SML	'>'	BIG
%left	RIGHT	LEFT
%left	'+'	'-'
%left	'*'	'/'	'%'
%right	'!'	'~'	INC	DEC
%left	DEFS
%left	'['	'('	MENBA	'.'	MAXM
%left	FUNC	SIZEOF
%left	CAST_DEF
%left	':'
%left	INT	FLOAT	DOUBLE
%left	LONG	SHORT	CHAR
%left	UNSIGNED	SIGNED

%%

program	: /* empty */
		{}
	| extdefs
		{}
	;

extdefs	: extdef
		{}
	| extdefs extdef
		{}
	;

extdef
	: clas_nul cast_nul data_list ';'
		{}
	| clas_nul cast_nul data_one '('
		{
		    register VALPTR *vp, *tp;
		    register CAST   *cp;

		    vp = (VALPTR *)($3);
		    cp = (CAST *)yysymb(sizeof(CAST));
		    cp->type = T_FUNC;
		    func_cp = cp->next = vp->cast;
		    cp->size = T_SIZE(T_FUNC);
		    vp->cast = cp;
		    vp->clas = C_CODE;

		    if ( (tp = get_cast(vp->name)) != NULL &&
			 	    tp->cast->type == T_FUNC ) {
			vp->addr = tp->addr;
		    	*((char **)(vp->addr)) = now_code_pos();
		    } else {
			vp->addr = yyheep(sizeof(char *));
		    	*((char **)(vp->addr)) = now_code_pos();
		    }

		    switch(cast_clas) {
		    case C_GLOB:
			glob_set(vp);
			break;
		    case C_LOCA:
			vp->next = loca_top;
			loca_top = vp;
			break;
		    }

		    stak_push();
		    stak_nst = sizeof(int) * (2 - 1);	/* epc & ebp */
		    stak_vac = 1;
		}
		param_list ')'
		{
		    stak_nst = 0;
		    stak_vac = (-1);
		}
		func_stmt
		{
		    gen_jmp_flush();
		    stak_pop();
		    stak_pos = HEEPMAX;
		}

	| error
		{ yyerror("Parse Syntax Error"); }
	;

func_stmt
	: '{'
		{
		    gen_op(CM_PUSH, T_POINT, 2);
		    gen_op(CM_BASE, T_POINT, 2);
		    gen_op(CM_ADDI, T_POINT, 4);
		    base_ptr = (int *)now_code_pos();
		    gen_addr(NULL);
		    func_ret = gen_label_make();
		}
		stmt_nul '}'
		{
		    gen_label_set(func_ret, now_code_pos());
		    *base_ptr = stak_nst;

		    gen_op(CM_ADDI, T_POINT, 4);
		    gen_addr((char *)(0 - stak_nst));
		    gen_op(CM_POP, T_POINT, 2);
		    gen_op(CM_RET, 0, 0);
		}
	| ';'
		{}
	;

param_list
	: param_def
	| param_list ',' param_def
	;

param_def
	: /* empty */
	| cast_nul { cast_clas = C_STAK; } data_ext
	;

clas_nul
	: /* empty */
		{ cast_clas = C_GLOB; }
	| STATIC
		{ cast_clas = C_LOCA; }
	;

cast_nul
	: CHAR
		{ cast_type = T_CHAR; }
	| SHORT
		{ cast_type = T_SHORT; }
	| LONG
		{ cast_type = T_LONG; }
	| INT
		{ cast_type = T_INT; }
	| UNSIGNED CHAR
		{ cast_type = T_UCHAR; }
	| UNSIGNED SHORT
		{ cast_type = T_USHORT; }
	| UNSIGNED LONG
		{ cast_type = T_ULONG; }
	| UNSIGNED INT
		{ cast_type = T_UINT; }
	| UNSIGNED
		{ cast_type = T_UINT; }
	| FLOAT
		{ cast_type = T_FLOAT; }
	| DOUBLE
		{ cast_type = T_DOUBLE; }
	;


data_list
	: data_def
	| data_list ',' data_def
	;

data_def
	: data_ext
	| data_ext '='
		{
		    register VALPTR *vp;
		    register CAST   *cp;
		    
		    vp = (VALPTR *)($1);
		    cast_ptr = vp->addr;
		    
		    if ( cast_clas == C_STAK ) {
			yyerror("Not Stack Data Init");
		    }
		    
		    if(vp->cast->type == T_ARRY) {
		      cast_ptr = *((char**)cast_ptr);
		    }
		    
		    
/*
		    cast_ptr = ((VALPTR *)($1))->addr;
*/
		}
		const_def
	;

const_def
	: const
		{
		    switch(cast_init) {
		    case T_CHAR:   case T_UCHAR:
			*((char *)cast_ptr) = $1;
			cast_ptr += sizeof(char);
			break;
		    case T_SHORT:  case T_USHORT:
			*((short *)cast_ptr) = $1;
			cast_ptr += sizeof(short);
			break;
		    case T_LONG:   case T_ULONG:
			*((long *)cast_ptr) = $1;
			cast_ptr += sizeof(long);
			break;
		    case T_FLOAT:
			*((float *)cast_ptr) = *((float *)($1));
			cast_ptr += sizeof(float);
			break;
		    case T_DOUBLE:
			*((double *)cast_ptr) = *((double *)($1));
			cast_ptr += sizeof(double);
			break;
/*
		    case T_FLOAT:  case T_DOUBLE:
			yyerror("Not Init Float or Double");
		        break;
*/
		    case T_POINT:
			*((int *)cast_ptr) = $1;
			cast_ptr += sizeof(int);
		        break;
		    }
		}
	| '{' const_list '}'
	;

const_list
	: const_def
	| const_list ',' const_def
	;

data_ext
	: data_one
		{
		    register VALPTR *vp;
/*		    register CAST   *cp;*/
/*		    register int    size;*/
		    register char   *ptr  /*,*ptr2*/  ;
/*		    register int    i,n,last_size,last_type;*/

		    vp = (VALPTR *)($1);
		    vp->clas = cast_clas;
		    switch(cast_clas) {
		    case C_GLOB:
			vp->addr = yyheep(vp->cast->size);
			
			if(vp->cast->type == T_ARRY) {
			  ptr = vp->addr;
			  vp->addr = yyheep(T_SIZE(T_INT));
			  *((char**)(vp->addr)) = ptr;
			}
			
			glob_set(vp);
			break;
		    case C_LOCA:
			vp->addr = yyheep(vp->cast->size);
			
			if(vp->cast->type == T_ARRY) {
			  ptr = vp->addr;
			  vp->addr = yyheep(T_SIZE(T_INT));
			  *((char**)(vp->addr)) = ptr;
			}
			
			stak_set(vp);
			break;
		    case C_STAK:
			stak_nst += (stak_vac * vp->cast->size);
		        stak_nst = alias(stak_nst);
			vp->addr = (char *)stak_nst;
			
			if(vp->cast->type == T_ARRY) {
			  ptr = vp->addr;
			  stak_nst += (stak_vac * T_SIZE(T_INT));
		          stak_nst = alias(stak_nst);
			  vp->addr = (char *)stak_nst;
			  
			  gen_op(CM_LEA, T_POINT, vp->clas);
			  gen_addr((char *)(ptr));
			  
			  gen_op(CM_EXG, T_POINT, 1);
			  
			  gen_op(CM_LEA, T_POINT, vp->clas);
			  gen_addr((char *)(vp->addr));
			  
			  gen_op(CM_STR, T_POINT, 1);
			}
			
			
			stak_set(vp);
			break;
		    }
		    $$ = $1;
		}
	;

data_one
	: LABEL
		{
		    register VALPTR *vp;

		    if ( cast_clas == C_STAK ) {
		        vp = (VALPTR *)yystak(sizeof(VALPTR));
		        vp->cast = (CAST *)yystak(sizeof(CAST));
		    } else {
		        vp = (VALPTR *)yysymb(sizeof(VALPTR));
		        vp->cast = (CAST *)yysymb(sizeof(CAST));
		    }
		    vp->name = (char *)($1);
		    vp->cast->type = cast_type;
		    vp->cast->next = NULL;
		    vp->cast->size = T_SIZE(cast_type);
		    cast_init = cast_type;
		    $$ = (int)vp;
		}
	| '*' data_one			%prec MAXM
		{
		    register VALPTR *vp;
		    register CAST   *cp;

		    if ( cast_clas == C_STAK )
		    	cp = (CAST *)yystak(sizeof(CAST));
		    else
		    	cp = (CAST *)yysymb(sizeof(CAST));

		    vp = (VALPTR *)($2);
		    cp->type = T_POINT;
		    cp->next = vp->cast;
		    cp->size = T_SIZE(T_POINT);
		    vp->cast = cp;
		    cast_init = T_POINT;
		    $$ = (int)vp;
		}
	| data_one '[' const ']'	%prec MAXM
		{
		    register VALPTR *vp;
		    register CAST   *cp;

		    if ( cast_clas == C_STAK )
		    	cp = (CAST *)yystak(sizeof(CAST));
		    else
		    	cp = (CAST *)yysymb(sizeof(CAST));

		    vp = (VALPTR *)($1);
		    cp->type = T_ARRY;
		    cp->next = vp->cast;
		    vp->cast = cp;
		    
		    while(1) {
		      cp->size = cp->next->size * $3;
		      if(cp->next->type != T_ARRY) break;
		      cp = cp->next;
		    }
		    $$ = (int)vp;
/*
		    vp = (VALPTR *)($1);
		    cp->type = T_ARRY;
		    cp->next = vp->cast;
		    cp->size = vp->cast->size * $3;
		    vp->cast = cp;
		    $$ = (int)vp;
*/
		}
	| data_one '[' ']'	%prec MAXM
		{
		    register VALPTR *vp;
		    register CAST   *cp;

		    if ( cast_clas == C_STAK )
		    	cp = (CAST *)yystak(sizeof(CAST));
		    else
		    	cp = (CAST *)yysymb(sizeof(CAST));

		    vp = (VALPTR *)($1);
		    cp->type = T_ARRY;
		    cp->next = vp->cast;
		    vp->cast = cp;
		    
		    while(1) {
		      cp->size = cp->next->size * 0;
		      if(cp->next->type != T_ARRY) break;
		      cp = cp->next;
		    }
		    $$ = (int)vp;
/*
		    vp = (VALPTR *)($1);
		    cp->type = T_ARRY;
		    cp->next = vp->cast;
		    cp->size = vp->cast->size * $3;
		    vp->cast = cp;
		    $$ = (int)vp;
*/
		}
	| '(' data_one ')' '(' ')'	%prec FUNC
		{
		    register VALPTR *vp;
		    register CAST   *cp;

		    if ( cast_clas == C_STAK )
		    	cp = (CAST *)yystak(sizeof(CAST));
		    else
		    	cp = (CAST *)yysymb(sizeof(CAST));

		    vp = (VALPTR *)($1);
		    cp->type = T_FUNC;
		    cp->next = vp->cast;
		    cp->size = T_SIZE(T_FUNC);
		    vp->cast = cp;
		    $$ = (int)vp;
		}
	| '(' data_one ')'		%prec MAXM
		{ $$ = $2; }
	;

data_type
	: /* empty */
		{
		    register VALPTR *vp;

		    vp = (VALPTR *)yystak(sizeof(VALPTR));
		    vp->cast = (CAST *)yystak(sizeof(CAST));
		    vp->name = NULL;
		    vp->cast->type = cast_type;
		    vp->cast->next = NULL;
		    vp->cast->size = T_SIZE(cast_type);
		    cast_init = cast_type;
		    $$ = (int)vp;
		}
	| '*' data_type			%prec MAXM
		{
		    register VALPTR *vp;
		    register CAST   *cp;

		    cp = (CAST *)yystak(sizeof(CAST));
		    vp = (VALPTR *)($2);
		    cp->type = T_POINT;
		    cp->next = vp->cast;
		    cp->size = T_SIZE(T_POINT);
		    vp->cast = cp;
		    cast_init = T_POINT;
		    $$ = (int)vp;
		}
	| data_type '[' const ']'	%prec MAXM
		{
		    register VALPTR *vp;
		    register CAST   *cp;

		    cp = (CAST *)yystak(sizeof(CAST));
		    vp = (VALPTR *)($1);
		    cp->type = T_ARRY;
		    cp->next = vp->cast;
		    cp->size = vp->cast->size * $3;
		    vp->cast = cp;
		    $$ = (int)vp;
		}
	| '(' data_type ')' '(' ')'	%prec FUNC
		{
		    register VALPTR *vp;
		    register CAST   *cp;

		    cp = (CAST *)yystak(sizeof(CAST));
		    vp = (VALPTR *)($1);
		    cp->type = T_FUNC;
		    cp->next = vp->cast;
		    cp->size = T_SIZE(T_FUNC);
		    vp->cast = cp;
		    $$ = (int)vp;
		}
	| '(' data_type ')'		%prec MAXM
		{ $$ = $2; }
	;

stmts
	: stmt
	| '{' { stak_push(); } stmt_nul '}' { stak_pop(); }
	;

stmt_nul
	: /* empty */
	| stmt_list
	;

stmt_list
	: stmts
	| local_def
	| stmt_list stmts
	| stmt_list local_def
	;

local_def
	: cast_nul { cast_clas = C_STAK; } data_list ';'
	| REGISTER cast_nul { cast_clas = C_STAK; } data_list ';'
	| STATIC cast_nul { cast_clas = C_LOCA; } data_list ';'
	;

stmt
	: gen_expr_nul ';'
		{}

	| if_head stmts		%prec IF
		{
		    gen_label_set($1, now_code_pos());
		}

	| if_head stmts ELSE
		{
		    $$ = gen_label_make();
		    gen_op(CM_JMP, 0, 0);
		    gen_jmp_make($$, now_code_pos());
		    gen_addr(NULL);

		    gen_label_set($1, now_code_pos());
		}
		stmts
		{
		    gen_label_set($4, now_code_pos());
		}

	| WHILE '('
		{
		    $$ = gen_label_make();
		    gen_label_set($$, now_code_pos());
		    break_push();									/* add */
		    gen_label_set(break_now() + 1, now_code_pos());	/* add */
		}
		test_expr ')'
		{
		    $$ = gen_label_make();
		    gen_op(CM_JE, 0, 0);
		    gen_jmp_make($$, now_code_pos());
		    gen_addr(NULL);
/*		    break_push();	delete */
		}
		stmts
		{
		    gen_op(CM_JMP, 0, 0);
		    gen_jmp_make($3, now_code_pos());
		    gen_addr(NULL);
		    gen_label_set($6, now_code_pos());
		    gen_label_set(break_pop(), now_code_pos());
		}

	| DO
		{
		    $$ = gen_label_make();
		    gen_label_set($$, now_code_pos());

		    break_push();
		    gen_label_set(break_now() + 1, now_code_pos());	/* add */
		}
		stmts WHILE '(' test_expr ')' ';'
		{
		    gen_op(CM_JNE, 0, 0);
		    gen_jmp_make($2, now_code_pos());
		    gen_addr(NULL);

		    gen_label_set(break_pop(), now_code_pos());
		}

	| FOR '(' gen_expr_nul ';'
		{
		    $$ = gen_label_make();
		    gen_label_set($$, now_code_pos());
		    break_push();									/* add */
		    gen_label_set(break_now() + 1, now_code_pos());	/* add */
		}
		test_expr_nul ';'
		{
		    $$ = gen_label_make();
		    gen_op($6 == FALSE ? CM_JMP:CM_JNE, 0, 0);
		    gen_jmp_make($$, now_code_pos());
		    gen_addr(NULL);

		    gen_label_make();		/* $$ + 1 */
		    gen_op(CM_JMP, 0, 0);
		    gen_jmp_make($$ + 1, now_code_pos());
		    gen_addr(NULL);

		    gen_label_make();		/* $$ + 2 */
		    gen_label_set($$ + 2, now_code_pos());
		}
		gen_expr_nul ')'
		{
		    gen_op(CM_JMP, 0, 0);
		    gen_jmp_make($5, now_code_pos());
		    gen_addr(NULL);

		    gen_label_set($8, now_code_pos());
	/*	    break_push();	delete	*/
		}
		stmts
		{
		    gen_op(CM_JMP, 0, 0);
		    gen_jmp_make($8 + 2, now_code_pos());
		    gen_addr(NULL);

		    gen_label_set($8 + 1, now_code_pos());
		    gen_label_set(break_pop(), now_code_pos());
		}

	| SWITCH '(' gen_expr ')' '{' 
		{
		    $$ = gen_label_make();
		    gen_op(CM_JMP, 0, 0);
		    gen_jmp_make($$, now_code_pos());
		    gen_addr(NULL);
		    break_push();
		    switch_make();
		}
		stmt_list '}'
		{
		    gen_op(CM_JMP, 0, 0);
		    gen_jmp_make(break_now(), now_code_pos());
		    gen_addr(NULL);

		    gen_label_set($6, now_code_pos());
		    switch_flush();

		    gen_label_set(break_pop(), now_code_pos());
		}

	| CASE const ':'
		{ case_make($2, now_code_pos()); }

	| DEFAULT ':'
		{ default_make(now_code_pos()); }

	| BREAK ';'
		{
		    gen_op(CM_JMP, 0, 0);
		    gen_jmp_make(break_now(), now_code_pos());
		    gen_addr(NULL);
		}
	| CONTINUE ';'
		{
		    gen_op(CM_JMP, 0, 0);
		    gen_jmp_make(break_now() + 1, now_code_pos());
		    gen_addr(NULL);
		}

	| RETURN expr ';'
		{
/*
printf("func=%ld\n",func_cp->type);
*/
		    gen_code((NODE *)($2));
		    gen_conv( now_cp->type , func_cp->type , 0 );
		    make_node_flush();
		    gen_op(CM_JMP, 0, 0);
		    gen_jmp_make(func_ret, now_code_pos());
		    gen_addr(NULL);
		}

	| LABEL ':'
		{ gen_label_set(goto_make((char *)$1), now_code_pos()); }

	| GOTO LABEL ';'
		{
		    gen_op(CM_JMP, 0, 0);
		    gen_jmp_make(goto_make((char *)$2), now_code_pos());
		    gen_addr(NULL);
		}
	;

if_head
	: IF '(' test_expr ')'
		{
		    $$ = gen_label_make();
		    gen_op(CM_JE, 0, 0);
		    gen_jmp_make($$, now_code_pos());
		    gen_addr(NULL);
		}
	;

const
	: const '&' const
		{ $$ = $1 & $3; }
	| const '|' const
		{ $$ = $1 | $3; }
	| const '^' const
		{ $$ = $1 ^ $3; }
	| const '+' const
		{ $$ = $1 + $3; }
	| const '-' const
		{ $$ = $1 - $3; }
	| const '*' const
		{ $$ = $1 * $3; }
	| const '/' const
		{ $$ = $1 / $3; }
	| const '%' const
		{ $$ = $1 % $3; }
	| const LEFT const
		{ $$ = $1 << $3; }
	| const RIGHT const
		{ $$ = $1 >> $3; }
	| '!' const
		{ $$ = !($2); }
	| '~' const
		{ $$ = ~($2); }
	| '-' const			%prec DEFS
		{ $$ = 0 - $2; }
	| '(' const ')'			%prec MAXM
		{ $$ = $2; }
	| '+' const			%prec DEFS
		{ $$ = $2; }
	| EXPR
		{ $$ = $1; }
	| DEXPR
		{ $$ = $1; }
	| STRING
		{ $$ = $1; }
	;

test_expr_nul
	: /* empty */
		{ $$ = FALSE; }
	| test_expr
		{ $$ = TRUE; }
	| test_expr_nul ',' test_expr
		{ $$ = TRUE; }
	;

test_expr
	: expr
		{
		    $$ = make_node('T', $1, 0);
		    gen_code((NODE *)($$));
		    make_node_flush();
		}
	;

gen_expr_nul
	: /* empty */
		{ $$ = FALSE; }
	| gen_expr
		{ $$ = TRUE; }
	| gen_expr_nul ',' gen_expr
		{ $$ = TRUE; }
	;

gen_expr
	: expr
		{
		    gen_code((NODE *)($1));
		    make_node_flush();
		}
	;

push_expr
	: /* empty */
		{ $$ = 0; }
	| expr
		{ $$ = make_node('p', 0, $1); }
	| push_expr ',' expr
		{ $$ = make_node('p', $1, $3); }
	;

expr
	: '(' cast_nul data_type ')' expr	%prec CAST_DEF
		{ $$ = make_node('@', $3, $5); }

	| SIZEOF '(' cast_nul data_type ')'
		{
		    register VALPTR *vp;

		    vp = (VALPTR *)($4);
		    $$ = make_node('E', vp->cast->size, 0);
		}
	| SIZEOF LABEL
		{
		    register VALPTR *vp;

		    if ( (vp = get_cast((char *)($2))) == NULL )
			yyerror("'%s' is undefine");
		    $$ = make_node('E', vp->cast->size, 0);
		}

	| expr '(' push_expr ')'
		{ $$ = make_node('c', $1, $3); }

	| expr ANDAND expr
		{ $$ = make_node('&', $1, $3); }
	| expr OROR expr
		{ $$ = make_node('|', $1, $3); }

	| expr '&' expr
		{ $$ = make_node((CM_AND << 8) | 'V', $1, $3); }
	| expr '|' expr
		{ $$ = make_node((CM_OR << 8) | 'V', $1, $3); }
	| expr '^' expr
		{ $$ = make_node((CM_XOR << 8) | 'V', $1, $3); }
	| expr '+' expr
		{ $$ = make_node((CM_ADD << 8) | 'V', $1, $3); }
	| expr '-' expr
		{ $$ = make_node((CM_SUB << 8) | 'V', $1, $3); }
	| expr '*' expr
		{ $$ = make_node((CM_MUL << 8) | 'V', $1, $3); }
	| expr '/' expr
		{ $$ = make_node((CM_DIV << 8) | 'V', $1, $3); }
	| expr '%' expr
		{ $$ = make_node((CM_MOD << 8) | 'V', $1, $3); }
	| expr LEFT expr
		{ $$ = make_node((CM_SHL << 8) | 'V', $1, $3); }
	| expr RIGHT expr
		{ $$ = make_node((CM_SHR << 8) | 'V', $1, $3); }

	| expr EQU expr
		{ $$ = make_node((CM_CE << 8) | 'x', $1, $3); }
	| expr UNEQU expr
		{ $$ = make_node((CM_CNE << 8) | 'x', $1, $3); }
	| expr '>' expr
		{ $$ = make_node((CM_CLT << 8) | 'x', $1, $3); }
	| expr '<' expr
		{ $$ = make_node((CM_CGT << 8) | 'x', $1, $3); }
	| expr BIG expr
		{ $$ = make_node((CM_CLE << 8) | 'x', $1, $3); }
	| expr SML expr
		{ $$ = make_node((CM_CGE << 8) | 'x', $1, $3); }

	| '!' expr
		{ $$ = make_node((CM_NOT << 8) | 'v', $2, 0); }
	| '~' expr
		{ $$ = make_node((CM_NOT << 8) | 'v', $2, 0); }
	| '-' expr			%prec DEFS
		{ $$ = make_node((CM_NEG << 8) | 'v', $2, 0); }

	| expr '?' expr ':' expr	%prec '?'
		{ $$ = make_node('?', $1, make_node(' ', $3, $5) ); }

	| '(' expr ')'			%prec MAXM
		{ $$ = $2; }

	| '+' expr			%prec DEFS
		{ $$ = $2; }

	| expr '=' expr
		{ $$ = make_node('=', $1, $3); }
	| expr ANDLET expr
		{ $$ = make_node((CM_AND << 8) | '=', $1, $3); }
	| expr ORLET expr
		{ $$ = make_node((CM_OR << 8) | '=', $1, $3); }
	| expr XORLET expr
		{ $$ = make_node((CM_XOR << 8) | '=', $1, $3); }
	| expr ADDLET expr
		{ $$ = make_node((CM_ADD << 8) | '=', $1, $3); }
	| expr SUBLET expr
		{ $$ = make_node((CM_SUB << 8) | '=', $1, $3); }
	| expr MULLET expr
		{ $$ = make_node((CM_MUL << 8) | '=', $1, $3); }
	| expr DIVLET expr
		{ $$ = make_node((CM_DIV << 8) | '=', $1, $3); }
	| expr MODLET expr
		{ $$ = make_node((CM_MOD << 8) | '=', $1, $3); }
	| expr LEFTLET expr
		{ $$ = make_node((CM_SHL << 8) | '=', $1, $3); }
	| expr RIGHTLET expr
		{ $$ = make_node((CM_SHR << 8) | '=', $1, $3); }

	| expr INC
		{ $$ = make_node((CM_INC << 8) | 'i', $1, 0); }
	| expr DEC
		{ $$ = make_node((CM_DEC << 8) | 'i', $1, 0); }
	| INC expr
		{ $$ = make_node((CM_INC << 8) | 'l', $2, 0); }
	| DEC expr
		{ $$ = make_node((CM_DEC << 8) | 'l', $2, 0); }

	| '*' expr			%prec DEFS
		{ $$ = make_node('I', $2, 0); }
	| '&' expr			%prec DEFS
		{ $$ = make_node('D', $2, 0); }

	| expr '[' expr ']'
		{ $$ = make_node('I',
		       make_node((CM_ADD << 8) | 'V', $1, $3), 0); }

	| EXPR
		{ $$ = make_node('E', $1, 0); }
	| DEXPR
		{ $$ = make_node('F', $1, 0); }
	| STRING
		{ $$ = make_node('S', $1, 0); }
	| LABEL
		{ $$ = make_label_node((char *)$1); }
	;

%%
