/*************************************************************************
*	׻
*************************************************************************/

#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>
#include	<setjmp.h>

#include	"expr.h"

#define	calc_error(_work,_err)	longjmp((_work)->jmp_buf,(_err))

typedef struct
{
	int			ch;
	char	   *buf;
	unsigned	pos;
	jmp_buf		jmp_buf;
} WORK;

static void _FAST	calc_next(WORK *work);
static int  _FAST	siki(WORK *work);
static int  _FAST	t1_siki(WORK *work);
static int  _FAST	t2_siki(WORK *work);
static int  _FAST	t3_siki(WORK *work);
static int  _FAST	t4_siki(WORK *work);
static int  _FAST	t5_siki(WORK *work);
static int  _FAST	calc_sign(WORK *work);
static int  _FAST	makedigit(WORK *work);
static int  _FAST	my_ishex(int c);
static int  _FAST	makehex(WORK *work);
static int  _FAST	makebin(WORK *work);
static int  _FAST	calc_num(WORK *work);

/*************************************************************************
*	 s ׻
*************************************************************************/

int _FAST	EXPR_calc( int *expr, const char *s )
{
	WORK	work;

	if ( expr == NULL || s == NULL )
		return (-1);
	*expr = 0;
	work.buf = NULL;
	work.pos = 0;

	{
		size_t	size;

		size = strlen(s) + 1;
		if ( (work.buf = malloc(size)) == NULL )
			return (-1);
		strcpy( work.buf, s);
#if	defined(CONFIG_USE_STRUPR)
		strupr( work.buf );
#else
		{
			char	*p;
			
			p = work.buf;
			while ( *p )
			{
				*p = toupper(*p);
				++p;
			}
		}
#endif
	}
	calc_next( &work );

	{
		int		ret;

		if ( (ret = setjmp(work.jmp_buf)) == 0 )
			*expr = siki(&work);
		free(work.buf);
		return ret;
	}
}

static void _FAST	calc_next(WORK *work)
{
	register int		ch;

	do
	{
		ch = work->buf[work->pos++];
	} while( ch == ' '|| ch == '\t' );
	work->ch = ch;
}

static int _FAST	siki(WORK *work)	/* ̤ͥϰ㤤	*/
{
	int 	temp;

	temp = t1_siki(work);
	for (;;)
	{
		switch ( work->ch )
		{
			case '^':	/* xor */
				calc_next(work);
				temp ^= t1_siki(work);
				break;
			default:
				return temp;
		}
	}
}

static int _FAST	t1_siki(WORK *work)
{
	int 	temp;

	temp = t2_siki(work);
	for (;;)
	{
		switch ( work->ch )
		{
			case '|':	/* or */
				calc_next(work);
				temp |= t2_siki(work);
				break;
			default:
				return	temp;
		}
	}
}

static int _FAST	t2_siki(WORK *work)
{
	int 	temp;

	temp = t3_siki(work);
	for (;;)
	{
		switch ( work->ch )
		{
			case '&':	/* and */
				calc_next(work);
				temp &= t3_siki(work);
				break;
			default:
				return	temp;
		}
	}
}

static int _FAST	t3_siki(WORK *work)
{
	int 	temp;

	temp = t4_siki(work);
	for (;;)
	{
		switch( work->ch )
		{
			case '<':
				if ( work->buf[work->pos] != '<' )
					calc_error(work,4);
				calc_next(work); calc_next(work);
				temp <<= t4_siki(work);
				break;
			case '>':
				if ( work->buf[work->pos] != '>' )
					calc_error(work,4);
				calc_next(work); calc_next(work);
				temp >>= t4_siki(work);
				break;
			default:
				return temp;
		}
	}
}

static int _FAST	t4_siki(WORK *work)
{
	int 	temp;

	temp = t5_siki(work);
	for (;;)
	{
		switch ( work->ch )
		{
			case '+':
				calc_next(work);
				temp += t5_siki(work);
				break;
			case '-':
				calc_next(work);
				temp -= t5_siki(work);
				break;
			default:
				return	temp;
		}
	}
}

static int _FAST	t5_siki(WORK *work)
{
	int 	temp, temp2;

	temp = calc_sign(work);
	for(;;)
	{
		switch ( work->ch )
		{
			case '*':
				calc_next(work);
				temp *= calc_sign(work);
				break;

			case '/':
				calc_next(work);
				temp2 = calc_sign(work);
				if ( temp2 == 0 )
					calc_error(work,5);
				temp /= temp2;
				break;

			case '%':
				calc_next(work);
				if ( temp2 == 0 )
					calc_error(work,5);
				temp2 = calc_sign(work);
				temp %= temp2;
				break;

			default:
				return temp;
		}
	}
}

static int _FAST	calc_sign(WORK *work)	/* ̤ͥϰֹ⤤	*/
{
	if ( work->ch == '-' )
	{
		calc_next(work);
		if ( work->ch == '~' )
		{
			calc_next(work);
			return ~(-calc_num(work));
		} else
			return (-calc_num(work));
	} else
	{
		if ( work->ch == '~' )
		{
			calc_next(work);
			return ~calc_num(work);
		} else
			return calc_num(work);
	}
}

static int _FAST	makedigit(WORK *work)
{
	int 	temp;

	temp = 0;
	while ( isdigit(work->ch) )
	{
		temp = temp*10 + (work->ch - '0');
		calc_next(work);
	}
	return	temp;
}

static int _FAST	my_ishex(int c)
{
	register int 	ch;

	if ( isdigit(c) )
		return (int)(c - '0');
	else if ( (ch = toupper(c)) >= 'A' && ch <= 'F' )
		return (int)(ch - 'A' + 10);
	else
		return (-1);
}

static int _FAST	makehex(WORK *work)
{
	int 	temp1, temp2;

	temp1 = 0;
	while ( (temp2 = my_ishex(work->ch)) >= 0 )
	{
		temp1 = (temp1<<4) + temp2;
		calc_next(work);
	}
	return	temp1;
}

static int _FAST	makebin(WORK *work)
{
	int 	temp1;

	temp1 = 0;
	for (;;)
	{
		if ( work->ch == '0' )
			temp1 = (temp1<<1);
		else if ( work->ch == '1' )
			temp1 = (temp1<<1) + 1;
		else
			break;
		calc_next(work);
	}
	return	temp1;
}


static int _FAST	calc_num(WORK *work)
{
	int 	temp;

	if ( (work->ch == '0' && toupper(work->buf[work->pos]) == 'X') )	/* "0x"	*/
	{
		calc_next(work);
		calc_next(work);
		return makehex(work);
	} else if ( (work->ch == '0' && toupper(work->buf[work->pos]) == 'B') )	/* "0b"	*/
	{
		calc_next(work);
		calc_next(work);
		return makebin(work);
	} else if ( work->ch == '\\' )
	{
		calc_next(work);
		temp = (work->ch & 0xFF);
		calc_next(work);
	} else if ( isdigit(work->ch) )
		return makedigit(work);
#if	0
	else if ( isalpha(work->ch) )
	{	/* ѿ	*/
		int		len = 0;
		char	name[32];

		name[0] = work->ch;
		while ( isalpha(work->ch) || isdigit(work->ch) || Calc_ch == '_' )
		{
			name[len++] = work->ch;
			calc_next();
			if ( len >= 32 )
				calc_error(work,3);	/* error */
		}
		name[len] = '\0';
		return (CALC_getval( name ));
	}
#endif
	else if ( work->ch == '(' )
	{
		calc_next(work);
		temp = siki(work);
		if( work->ch == ')' )
			calc_next(work);
		else
			calc_error(work,2);
	} else
		calc_error(work,3);

	return temp;
}

#if	0
typedef struct _var_t
{
	struct _var_t	*next;
	char			name[32];
	int				val;
} VAR_T;

#define	HASH_MAX	(32)

static VAR_T	*HashTbl[HASH_MAX] =
{
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
};

static int		HASH( const char *name )
{
	register int		hash = 0;

	while ( *name )
	{
		hash = hash + (toupper(*name)) * 31;
		++name;
	}
	return (hash & (HASH_MAX-1));
}

int		CALC_let( const char *name, int val )
{
	int		hash = HASH(name);
	VAR_T	*var = HashTbl[hash];

	while ( var )
	{
		if ( strcmpi( var->name, name) == 0 )
			break;
		var = var->next;
	}
	if ( var )
	{
		var->val = val;
	} else
	{
		if ( (var = calloc(sizeof(VAR_T),1)) == NULL )
			return (-1);
		strncpy( var->name, name, 31 );
		var->val = val;

		var->next     = HashTbl[hash];
		HashTbl[hash] = var;
	}
	return (0);
}

int		CALC_getval( CONST char *name )
{
	int		hash = HASH(name);
	VAR_T	*var = HashTbl[hash];

	while ( var )
	{
		if ( strcmpi( var->name, name) == 0 )
			return (var->val);
		var = var->next;
	}
	return (0);
}

void	CALC_free(void)
{
	int		hash;

	for ( hash = 0; hash < HASH_MAX; ++hash )
	{
		VAR_T	*var, *next;

		var = HashTbl[hash];
		while ( var )
		{
			next = var->next;
			free(var);
			var = next;
		}
		HashTbl[hash] = NULL;
	}
}

void	CALC_clear(void)
{
	int		hash;

	for ( hash = 0; hash < HASH_MAX; ++hash )
	{
		VAR_T	*var;

		var = HashTbl[hash];
		while ( var )
		{
			var->val = 0;
			var = var->next;
		}
	}
}

int		CALC_atoi( CONST char *s )
{
	int		val;

	if ( CALC_exe(&val,s) == 0 )
		return (val);
	else
		return (-9999);
}
#endif
