/*************************************************************************
*	"he386.c" -- compiler main loop
*************************************************************************/

#include	<stdio.h>
#include	<stdlib.h>
#include	<stdarg.h>
#include	<ctype.h>
#include	<string.h>

#include	"he386.h"
#include	"mmoption.h"
#include	"mmfile.h"
#include	"mml.h"
#include	"mmac.h"
#include	"mmpp.h"
#include	"mmlib.h"
#include	"mmpart.h"
#include	"mmtrk.h"
#include	"mmptn.h"
#include	"mmctrl.h"
#include	"mmdec.h"
#include	"mmcmdtyp.h"
#include	"mmlist.h"
#include	"mmsg.h"
#include	"msgdat.h"
#include	"midicc.h"
#include	"mmcci.h"
#include	"cci.h"

#ifdef	__HIGHC__
#	pragma	On(Align_labels);
#endif

/*	config	------------------------------------------------------------*/
/*	#define	_DEBUG	*/
/*--------------------------------------------------------------------*/

MMSQR_T	MMsqr;

static int  _FAST	get_cci_filename(char *fn);
static void _FAST	set_strcolor( int col );
static void _FAST	set_baccolor( int col );

void	HE_puts(MMSQR_T *sqr, int mode, CONST char *s)
{
	static MMSQR_T	   *m_sqr = NULL;
	static unsigned		length = 0;

	if ( sqr == NULL )
	{
		if ( (sqr = m_sqr) == NULL )
			sqr = &MMsqr;
	}
	switch( mode )
	{
		case _HE_MSG_INF:
		case _HE_MSG_MPP:
		case _HE_MSG_CTR:
		case _HE_MSG_REM:
		case _HE_MSG_INP:
		case _HE_MSG_OUT:
		case _HE_MSG_TXT:
			if ( sqr->infFp != NULL && sqr->infFp != stderr && sqr->infFp != stdout )
			{	/* ե	*/
				fputs(  s  , sqr->infFp );
				fputs( "\n", sqr->infFp);
			}
			if ( mode == _HE_MSG_INF || sqr->output.cns )
			{
				if ( sqr->msgMode == 0 )
					USR_puts( "\n" );
				USR_puts( s );
				USR_puts( "\n" );
				length = 0;
				sqr->msgMode = 1;
			}
			break;

		case _HE_MSG_LIN:
#if	0
			{
				USR_putsStderr( "\r\x1BT");	/* ֤ԤνޤǺ	*/
				length = strlen(s);
				USR_putsStderr( s );
				USR_putsStderr("\r");
			}
#else
			{
				size_t	len;

				len = strlen(s);
				USR_putcStderr('\r');
				USR_putsStderr(s);
				if ( length > len )
				{	/* ޤǺ	*/
					char		tmp[256];

					memset( tmp, ' ', length - len);
					tmp[length - len    ] = '\r';
					tmp[length - len + 1] = '\0';
					USR_putsStderr(tmp);
				}
				length = len;
			}
#endif
			sqr->msgMode = 0;
			break;

		case _HE_MSG_ERR:
			if ( sqr->errFp != NULL && sqr->errFp != stderr && sqr->errFp != stdout )
			{
				fputs(  s  , sqr->errFp );	/* ե	*/
				fputs( "\n", sqr->errFp);
			}
			if ( sqr->msgMode == 0 )
				USR_puts( "\n");
			USR_puts( s );
			USR_puts( "\n");
			length = 0;
			sqr->msgMode = 1;
			break;

		case _HE_MSG_DEB:
		case _HE_MSG_DE1:
		case _HE_MSG_DE2:
		case _HE_MSG_DE3:
		case _HE_MSG_DE4:
		{
			static char	tbl[4] = {C_GREEN,C_YELLOW,C_CYAN,C_WHITE};

			if ( !sqr->output.deb )
				break;
			if ( sqr->msgMode == 0 )
				USR_puts( "\n" );

			set_strcolor( tbl[mode - _HE_MSG_DE1] );
			set_baccolor(C_BLACK);
			USR_puts( s );
			set_strcolor(ERR);
			USR_puts( "\n" );
			length = 0;
			sqr->msgMode = 1;
			break;
		}
	}
	m_sqr = sqr;
}

static void _FAST	set_strcolor(int col)
{
	static char	coltbl[16] = {30,34,31,35,32,36,33,37};

	if ( col < 0 )
		USR_puts("\x1B[0m");
	else
		USR_fprintf(stderr,"\x1B[%dm", coltbl[col & 7]);
}

static void _FAST	set_baccolor(int col)
{
	static char	coltbl[16] = {40,44,41,45,42,46,43,47};

	if ( col < 0 )
		USR_puts("\x1B[0m");
	else
		USR_fprintf(stderr,"\x1B[%dm", coltbl[col & 7] );
}

void	HE_printf(MMSQR_T    *sqr, int mode, CONST char *form, ...)
{
	va_list		arg;
	static char	tmp[_MM_DEC_BUFSIZ+4];

	va_start( arg, form );
	vsprintf( tmp, form, arg );
	va_end(arg);

	HE_puts( sqr, mode, tmp );
}

/*************************************************************************
*	ե륯
*************************************************************************/

void	he386_fclose_all(void)
{
	MMSQR_T	   *sqr = &MMsqr;

	MMpp_close( sqr );
	if ( sqr->inpFp )
	{
		FS_fclose(sqr->inpFp);	sqr->inpFp = NULL;
	}
	if ( sqr->outFp )
	{
		FS_fclose(sqr->outFp);	sqr->outFp = NULL;
	}
	if ( sqr->errFp )
	{
		FS_fclose(sqr->errFp);	sqr->errFp = NULL;
	}
	if ( sqr->infFp )
	{
		FS_fclose(sqr->infFp);	sqr->infFp = NULL;
	}
}

/*************************************************************************
*	ǽѡuserbreak()ƤФ
*************************************************************************/
void	he386_break(void)
{
	he386_fclose_all();
	he386_term();
}

int		he386_init(void)
{
	MMSQR_T	   *sqr = &MMsqr;

	if ( sqr->errFn[0] )
		sqr->errFp = FS_fopen(sqr->errFn, FOPEN_WRITE_TXT);

	if ( sqr->infFn[0] )
		sqr->infFp = FS_fopen(sqr->infFn, FOPEN_WRITE_TXT);


	{	/* CCI 	*/
		extern int			CodeSize;
		extern int			HeepSize;

#ifdef	_DEBUG
		fprintf(stderr, "CodeSize = %d\n", CodeSize);
		fprintf(stderr, "HeepSize = %d\n", HeepSize);
#endif
		if ( CCI_init( CodeSize, HeepSize) )
			return (ERR);

		MM_cci_setFunc();
		{
			char	fn[256];

			if ( CciPath[0] == '\0' )
				strcpy( CciPath, StartPath );
			strcpy( fn, CciPath);
			if ( get_cci_filename(fn) ==  0 )
			{
				sprintf( sqr->errMsgBuf, _msg_he386_cci_program_file_not_found[TypMsg], CciFileName);
				sqr->errMsg = sqr->errMsgBuf;
				return (ERR);
			}
#ifdef	_DEBUG
			printf("CCI Filename = %s\n", fn);
#endif
			if ( CCI_load( fn ) )	/* CCI ץɤ߹	*/
			{
				sqr->errMsg = CCI_getErrMsg();
				return (ERR);
			}
		}
	}

	mm_srand(0);	/* 	*/

	{	/* MML ޥɥơ֥ν	*/
		int		i;

		static int		tbl[26] =
		{
			/* A */ -1,
			/* B */ -1,
			/* C */ -1,
			/* D */ -1,
			/* E */ -1,
			/* F */ -1,
			/* G */ -1,
			/* H */ MMCMDTYP_CH,
			/* I */ 0,
			/* J */ 0,
			/* K */ MMCMDTYP_KEYSHIFT,
			/* L */ -1,
			/* M */ MMCMDTYP_CCV + MID_CC_MODURATION,
			/* N */ -1,
			/* O */ MMCMDTYP_OCTAVE,
			/* P */ MMCMDTYP_CCV + MID_CC_PANPOT,
			/* Q */ MMCMDTYP_GATERATIO,
			/* R */ -1,
			/* S */ MMCMDTYP_CLOCK,
			/* T */ MMCMDTYP_TEMPO,
			/* U */ MMCMDTYP_PITCH,
			/* V */ MMCMDTYP_VELOCITY,
			/* W */ MMCMDTYP_CCV + MID_CC_VOLUME,
			/* X */ MMCMDTYP_CCV + MID_CC_EXPRESSION,
			/* Y */ 0,
			/* Z */ 0,
		};
		for ( i = 0; i < 26; ++i )
			sqr->mmlCmd[i] = tbl[i];
	}

	sqr->end_flag     = FALSE;
	sqr->division     = MM_DEFAULT_DIVISION;
	sqr->tempoSetFlag = FALSE;
	sqr->top_tempo    = 60000000 / 120;
	sqr->tempo        = 60000000 / 120;

	sqr->signSetFlag = FALSE;
	sqr->top_sign[0] = 4;
	sqr->top_sign[1] = 2;				/* 2   */
	sqr->top_sign[2] = 0x18;
	sqr->top_sign[3] = 8;
	sqr->sign[0]     = 4;
	sqr->sign[1]     = 2;
	sqr->sign[2]     = 0x18;
	sqr->sign[3]     = 8;

	sqr->clk = sqr->max_clk = sqr->meas = 0;

	sqr->title[0]     = '\0';
	sqr->copyright[0] = '\0';

	sqr->errMsgBuf[0] = '\0';
	sqr->errMsg       = sqr->errMsgBuf;
	sqr->workSize     = 0;

	/* ޥطν	*/
	if ( MMac_init( sqr ) )			return (ERR);

	/* command line option */
	{
		LISTDATA	*listdata;

		for ( listdata = ListMac.datTop; listdata != NULL; listdata = listdata->next )
		{
			const char	   *s;
			char			name[256];

			s = listdata->ptr;
			if ( mm_getname(name, sizeof(name)-1, &s) < 0 )
				break;
			if ( *s == '=' )
				++s;
			MMac_extDefine(sqr, name, s, MMAC_ATT_NORMAL, (int (*)())0);
		}
	}

	/* ȥåν	*/
	if ( MMpart_init( sqr ) )		return (ERR);
	if ( MMtrk_init( sqr )  )		return (ERR);
	if ( MMptn_init( sqr )  )		return (ERR);

	/* CCI ؿν	*/
	CCI_exec("HE386_start",1,sqr);

	/* ץץåΥץ	*/
	FS_appendSnam( sqr->inpFn, sqr->inpFn, _SNAM_MML );
	HE_printf( sqr, _HE_MSG_INF, _msg_he386_source_filename[TypMsg], sqr->inpFn );
	if ( MMpp_open(sqr,sqr->inpFn) )
		return (-1);

	return (0);
}

void	he386_term(void)
{
	MMSQR_T	   *sqr = &MMsqr;

	MMpp_close(sqr);
	MMptn_term(sqr);
	MMtrk_term(sqr);
	MMpart_term(sqr);
	MMac_term(sqr);
}

void	he386_error(void)
{
	MMSQR_T		   *sqr = &MMsqr;
	MMINC_T		   *mminc;
	extern void		userbreak(int level);

	if ( sqr->incLevel > 0 && sqr->incLevel < _MMINC_MAX )
	{
		mminc = &(sqr->inc[sqr->incLevel-1]);
		HE_printf( sqr, _HE_MSG_ERR, "%s (%d) : %s", mminc->fn, mminc->ln, sqr->errMsg );
	} else
		HE_printf( sqr, _HE_MSG_ERR, "%s", sqr->errMsg );
	if ( sqr->errPoint )
		HE_printf( sqr, _HE_MSG_ERR, _msg_he386_error_point[TypMsg], sqr->errPoint );

	userbreak(1);
}

int		he386_main(void)
{
	MMSQR_T	   *sqr = &MMsqr;
	time_t		tm_start, tm_end;

	HE_printf( sqr, _HE_MSG_INF,
#if	defined(__HE386__)
		"\n386|High-EUP Compiler ver." _MAIN_VERSION " edition " _MAIN_EDITION "\n"
#endif
#if	defined(__HE32__)
		"High-EUP Compiler ver." _MAIN_VERSION " edition " _MAIN_EDITION "\n"
#endif
	);
	HE_printf( sqr, _HE_MSG_DE3, "mmtrk  work size  %8u", sizeof(MMTRK_T) );
	HE_printf( sqr, _HE_MSG_DE3, "mmpart work size  %8u", sizeof(MMPART_T));

	if ( he386_init() )
	{
		if ( sqr->errMsg )
			HE_puts( sqr, _HE_MSG_ERR, sqr->errMsg  );
		he386_fclose_all();
		return (ERR);
	}

	tm_start = time(NULL);
	HE_printf( sqr, _HE_MSG_INF, _msg_he386_compile_start[TypMsg], mm_getTimeStr(tm_start) );
	HE_puts( sqr, _HE_MSG_INF, "" );

	while ( sqr->end_flag == FALSE )
	{
		char	   *s;

		sqr->errMsg = sqr->errPoint = NULL;
		if ( (s = MMpp_getline(sqr)) == NULL )
		{	/* normal end */
			if ( sqr->errMsg )
				he386_error();
			if ( sqr->end_flag == FALSE )
			{
				if ( MMctr_end( sqr, "" ) )
					he386_error();
			}
			break;
		}
		if ( *s )
		{
			if ( *s == '/' )
			{	/* 	*/
				HE_puts( sqr, _HE_MSG_REM, s );
			} else if ( *s == '.' )
			{	/* ȥ	*/
				if ( MMctr_line( sqr, s ) )
					he386_error();
			} else
			{	/* MML */
				if ( MML_line( sqr, s ) )
					he386_error();
			}
		}
	}

	HE_puts( sqr, _HE_MSG_INF, "" );

	{
		long		tm_dif;
		char		tmp[64];

		tm_end = time(NULL);
		tm_dif = (long)difftime(tm_end,tm_start);
		if ( tm_dif > 60 )
			sprintf( tmp, _msg_he386_tm_msg1[TypMsg], tm_dif, tm_dif/60l, tm_dif % 60l );
		else
			sprintf( tmp, _msg_he386_tm_msg2[TypMsg], tm_dif );
		HE_printf( sqr, _HE_MSG_INF, _msg_he386_compile_end[TypMsg], mm_getTimeStr(tm_end) );
		HE_printf( sqr, _HE_MSG_INF, _msg_he386_compile_time[TypMsg], tmp );
	}

	/* ׽	*/
	if ( sqr->output.dmp )
		MMlist_output( sqr );

	he386_fclose_all();
	he386_term();

	return (NORMAL);
}

extern int		MM_optimize( MMSQR_T *sqr );
extern int		MMsmf_generate( MMSQR_T *sqr );

int		he386_genarate( MMSQR_T *sqr, MMCLK_T clk )
{
	int		i;

	for ( i = 0; i < _MMSQR_TRK_HASH_MAX; ++i )
	{
		MMTRK_T	   *trk;

		for ( trk = sqr->trkTbl[i]; trk != NULL; trk = trk->next )
			mmtrk_set_end( trk, clk );
	}
	sqr->end_flag = TRUE;

	/* γ	*/
	for ( i = 0; i < _MMSQR_TRK_HASH_MAX; ++i )
	{
		MMTRK_T	   *trk;

		for ( trk = sqr->trkTbl[i]; trk != NULL; trk = trk->next )
			MMtrk_workFree( sqr, trk);
	}
	MMptn_term( sqr );	/* ѥǡ	*/
	MMac_term( sqr );	/* ޥǡ		*/

	{
		char	tmp[32];

		HE_printf( sqr, _HE_MSG_INF, _msg_he386_using_worksize[TypMsg], GetSizeStr( sqr->workSize, tmp ) );
	}

	/* ץƥޥ	*/
	if ( MM_optimize( sqr ) )
		return (ERR);

	return MMsmf_generate( sqr );
}


static int _FAST	get_cci_filename(char *fn)
{
	FILE				   *fp;

	FS_addFnam(fn, CciFileName);
	if ( (fp = FS_fopen(fn, "r")) != NULL )
	{
		FS_fclose(fp);
		return (1);	/* OK */
	}

	{
		char				   *env, *tok, tmp[1024];
		static const char		sep[] = COMMAND_PATH_SEP;

		if ( (env = getenv(COMMAND_PATH_ENV)) == NULL )
			return (0);	/* false */
		strncpy( tmp, env, sizeof(tmp));
		tok = strtok(tmp, sep);
		while ( tok != NULL )
		{
			strcpy( fn, tok);
			FS_addFnam(fn, CciFileName);
			if ( (fp = FS_fopen(fn,"r")) != NULL )
			{
				FS_fclose(fp);
				return (1);	/* OK */
			}
			tok = strtok(NULL, sep);
		}
		
	}

	return (0);	/* false */
}
