/*************************************************************************
*	
*************************************************************************/

#include	<stdio.h>
#include	<stdlib.h>
#include	<ctype.h>
#include	<string.h>
#include	<math.h>

#include	"he386.h"
#include	"msgdat.h"
#include	"mml.h"
#include	"mmlib.h"
#include	"mmpp.h"
#include	"mmac.h"

/*	config	------------------------------------------------------------*/
/*	#define	_DEBUG	*/
/*--------------------------------------------------------------------*/

static int _FAST	mm_get_steptime( MMSQR_T *sqr, CONST char *s );

#define	ARG_MAX		(32)
#define	ARG_BUFSIZ	(32)

/*--------------------------------------------------------------------*/
#if	defined(WIN32)
#	pragma	pack(push,1)
#endif
typedef	struct
{
	const char	   *name;
	char			n;
	char			note[7];
} chord_t;
#if	defined(WIN32)
#	pragma	pack(pop)
#endif

chord_t	CodeTbl[] =
{
	{	"",		3,	{ 4, 7, 0, 0, 0, 0, 0}	},
	{	"+11",	1,	{ 4, 7, 9,10,14,18, 0}	},	/* 㡼ץɡ֥				*/
	{	"+5",	3,	{ 4, 8, 0, 0, 0, 0, 0}	},	/* 							*/
	{	"-5",	3,	{ 4, 6, 0, 0, 0, 0, 0}	},	/* 륿								*/
	{	"11",	1,	{ 4, 7,10,14,17, 0, 0}	},	/* ֥							*/
	{	"13",	1,	{ 4, 7,10,14,17,21, 0}	},	/* ƥ						*/
	{	"13#11",1,	{ 4, 7,10,11,14,18,21}	},	/* ƥ󥹡㡼ץɡ֥	*/
	{	"13(#11)",1,{ 4, 7,10,11,14,18,21}	},	/* ƥ󥹡㡼ץɡ֥	*/
	{	"13(+11)",1,{ 4, 7,10,11,14,18,21}	},	/* ƥ󥹡㡼ץɡ֥	*/
	{	"13+11",1,	{ 4, 7,10,11,14,18,21}	},	/* ƥ󥹡㡼ץɡ֥	*/
	{	"6",	4,	{ 4, 7, 9, 0, 0, 0, 0}	},	/* å								*/
	{	"69",	6,	{ 4, 7, 9,14, 0, 0, 0}	},	/* åʥ					*/
	{	"7",	4,	{ 4, 7,10, 0, 0, 0, 0}	},	/* ֥								*/
	{	"7+5",	4,	{ 4, 8,10, 0, 0, 0, 0}	},	/* ȡ֥				*/
	{	"7+9",	1,	{ 4, 7,10,15, 0, 0, 0}	},	/* ֥󡦥㡼ץɡʥ			*/
	{	"7-5",	4,	{ 4, 6,10, 0, 0, 0, 0}	},	/* 륿ɡ֥						*/
	{	"7-9",	4,	{ 4, 7,10,13, 0, 0, 0}	},	/* ֥󡦥եåƥåɡʥ			*/
	{	"7sus4",4,	{ 5, 7,10, 0, 0, 0, 0}	},	/* ֥󥹡ڥǥåɡե		*/
	{	"9",	4,	{ 4, 7,10,14, 0, 0, 0}	},	/* ʥ								*/
	{	"9+5",	4,	{ 4, 8,10,14, 0, 0, 0}	},	/* ȡʥ				*/
	{	"9-5",	4,	{ 4, 6,10,14, 0, 0, 0}	},	/* 륿ɡʥ						*/
	{	"9alt",	4,	{ 4, 6,10,14, 0, 0, 0}	},	/* 륿ɡʥ						*/
	{	"9aug",	4,	{ 4, 8,10,14, 0, 0, 0}	},	/* ȡʥ				*/
	{	"alt",	3,	{ 4, 6, 0, 0, 0, 0, 0}	},	/* 륿								*/
	{	"alt7",	4,	{ 4, 6,10, 0, 0, 0, 0}	},	/* 륿ɡ֥						*/
	{	"aug",	3,	{ 4, 8, 0, 0, 0, 0, 0}	},	/* 							*/
	{	"aug7",	4,	{ 4, 8,10, 0, 0, 0, 0}	},	/* ȡ֥				*/
	{	"dim7",	4,	{ 3, 6, 9, 0, 0, 0, 0}	},	/* ǥߥ˥åɡ֥			*/
	{	"dim9",	1,	{ 3, 6, 9,14, 0, 0, 0}	},	/* ǥߥ˥åɡʥ			*/
	{	"dimj7",1,	{ 3, 6,11, 0, 0, 0, 0}	},	/* ǥߥ˥åɡ᥸㡼֥	*/
	{	"m",	3,	{ 3, 7, 0, 0, 0, 0, 0}	},	/* ޥʡ									*/
	{	"m11",	1,	{ 3, 7,10,14,17, 0, 0}	},	/* ޥʡ֥						*/
	{	"m11-5",1,	{ 3, 6,10,14,17, 0, 0}	},	/* 륿ɡޥʡ֥			*/
	{	"m13",	1,	{ 3, 7,10,14,17,21, 0}	},	/* ޥʡƥ					*/
	{	"m6",	4,	{ 3, 7, 9, 0, 0, 0, 0}	},	/* ޥʡå						*/
	{	"m69",	6,	{ 3, 7, 9,14, 0, 0, 0}	},	/* ޥʡåʥ				*/
	{	"m7",	4,	{ 3, 7,10, 0, 0, 0, 0}	},	/* ޥʡ֥						*/
	{	"m7-5",	4,	{ 3, 6,10, 0, 0, 0, 0}	},	/* 륿ɡޥʡ֥			*/
	{	"m9",	4,	{ 3, 7,10,14, 0, 0, 0}	},	/* ޥʡʥ						*/
	{	"m9-5",1,	{ 3, 6,10,14, 0, 0, 0}	},	/* 륿ɡޥʡ֥			*/
	{	"Maj7",	4,	{ 4, 7,11, 0, 0, 0, 0}	},	/* ᥸㡼֥						*/
	{	"Maj9",	4,	{ 4, 7,11,14, 0, 0, 0}	},	/* ᥸㡼ʥ						*/
	{	"mMaj7",4,	{ 3, 7,11, 0, 0, 0, 0}	},	/* ޥʡ᥸㡼֥				*/
	{	"mMaj9",4,	{ 3, 7,11,14, 0, 0, 0}	},	/* ޥʡ᥸㡼ʥ				*/
	{	"sus4",	3,	{ 5, 7, 0, 0, 0, 0, 0}	},	/* ڥǥåɡե					*/
};

static int		compfunc( const chord_t *key1, const chord_t *key2)
{
	return (strcmpi(key1->name, key2->name));
}

int		CodeGenerate( int note[8], CONST char *chordName, int sp )
{
	const char	   *s;
	int				base, i, j, nn;
	chord_t		   *chord;

	for ( i = 0; i < 8; ++i )
		note[i] = -1;

	s = chordName;
	while ( isspace(*s) )
		++s;
	if ( toupper(*s) < 'A' || toupper(*s) > 'G' )
		return (ERR);
	{
		static	int	tbl[7] = {9,11,0,2,4,5,7};
		note[0] = 48 + tbl[ toupper(*s) - 'A' ];
	}
	++s;
	if ( *s == '+' || *s == '#' )
	{
		++s;
		++note[0];
	} else if ( *s == '-' )
	{
		++s;
		--note[0];
	}

	while ( isspace(*s) )
		++s;

	/*	̾	*/
#if	1
	{
		chord_t		key;

		key.name = s;
		if ( (chord = bsearch( &key, CodeTbl, sizeof(CodeTbl)/sizeof(*CodeTbl),
		  sizeof(*CodeTbl), (int (*)(const void *, const void *))compfunc)) == NULL )
			return (-1);
	}
#else
	for ( i = 0; ; ++i )
	{
		chord = &CodeTbl[i];
		if ( chord->name == NULL )
			return (ERR);
		if ( strcmpi( s, chord->name) ==  0 )
			break;
	}
#endif

	/* ɹ	*/
	nn = 1;
	for ( nn = 1, i = 0; chord->note[i] != 0 && i < 7; ++i, ++nn )
		note[nn] = note[0] + chord->note[i];

	base = note[0];
	if ( sp > 0 && chord->n > 1 )
	{	/* ž	*/
		sp %= chord->n;
		if ( chord->n == 6 && sp >= 4 )
		{	i = note[2];
			note[2] = note[3];
			note[3] = note[4];
			note[4] = i + 12;
			if ( sp == 5 )
			{	note[5] = note[0] + 24;
			}
		} else
		{	while ( sp-- > 0 )
			{	if ( chord->n == nn )
				{	/* ž	*/
					i = note[0];
					for ( j = 0; j < nn - 1; ++j )
						note[j] = note[j+1];
					note[nn-1] = i + 12;
				} else
				{	/* 롼ȲѲʤ	*/
					i = note[1];
					for ( j = 1; j < nn - 1; ++j )
						note[j] = note[j+1];
					note[nn-1] = i + 12;
				}
			}
		}
	}
	return (base);
}

/*************************************************************************
*	$chord( chord, spin, [step, gate [, tie]] )
*************************************************************************/

int		MMfunc_chord(MMSQR_T *sqr,size_t bufSiz, char *buf,int argc, char **argv)
{
	int			i, ret = NORMAL;
	int			spin;
	char		gate_buf[32];
	CONST char	*chord, *step, *gate, *tie;
	int			dl = 0;
	char		*dp, *sbuf = NULL;

#define	term()	{	if ( sbuf )	free( sbuf );}

	if ( argc < 1 )
	{
_PARA_ERR:
		sqr->errMsg = "Code parameter error!";
		term();
		return (ERR);
	}
	if ( (sbuf = malloc(_MM_DEC_BUFSIZ)) == NULL )
	{
		sqr->errMsg = _OUT_OF_MEMORY;
		term();
		return (ERR);
	}
	dp = sbuf;

	chord = argv[0];
	spin = argc > 1 ? atoi(argv[1]) : 0;
	step = argc > 2 ? argv[2] : ":0";
	if ( argc > 3 )
		gate = argv[3];
	else
	{
		if ( argc > 2 )
		{
			int			gt;

			if ( (gt = mm_get_steptime( sqr, argv[2] )) < 0 )
				goto _PARA_ERR;
			if ( gt > 0 )
				--gt;
			sprintf( gate_buf, "%d", gt );
			gate = gate_buf;
		} else
			gate = "0";
		
	}
	tie  = argc > 4 ? argv[4] : "";

	{
		int		base, note[8], b_oct;

		if ( (base = CodeGenerate( note, chord, spin )) < 0 )
			goto _PARA_ERR;
#if	0
		b_oct = base / 12;
#else
		b_oct = note[0] / 12;
#endif
		i = 0;
		while ( i < 8 )
		{
			if ( note[i] < 0 )
				break;
			{
				int		oct;
				char	up[16], tmp[256], down[16];

				oct = note[i]/12;
				if ( oct > b_oct )
				{
					int		j;
					char	*s;

					s = up;
					for ( j = b_oct; j < oct; ++j )
						*s++ = '>';
					*s = '\0';

					s = down;
					for ( j = b_oct; j < oct; ++j )
						*s++ = '<';
					*s = '\0';
				} else if ( oct < b_oct )
				{
					int		j;
					char	*s;

					s = up;
					for ( j = b_oct; j < oct; ++j )
						*s++ = '<';
					*s = '\0';

					s = down;
					for ( j = b_oct; j < oct; ++j )
						*s++ = '>';
					*s = '\0';
				} else
					up[0] = down[0] = '\0';

				{
					int		l;
					static CONST char	*tbl[12] =
					{
						"C","C+","D","D+","E","F","F+","G ","G+","A ","A+","B"
					};

					sprintf( tmp, "%s%s:0,%s%s%s", up, tbl[note[i]%12], gate, tie, down);
					l = strlen( tmp );
					if ( (unsigned)(dl + l) >= bufSiz )
						goto _SET_ERR;
					strcpy( dp, tmp );
					dl += l;
					dp += l;
				}
			}
			++i;
		}
	}

	{
		char		tmp[256];
		unsigned	l;

		sprintf(tmp, "R%s", step );
		l = strlen( tmp );
		if ( (unsigned)dl + l >= bufSiz )
			goto _SET_ERR;
		strcpy( dp, tmp );
		dl += l;
		dp += l;
	}

	if ( (unsigned)dl >= bufSiz )
	{
_SET_ERR:
		sqr->errMsg = _OUT_OF_STRING_SPACE;
		ret = ERR;
	} else
	{
		memcpy( buf, sbuf, dl );
		ret = dl;
	}
/*_RET:*/
	term();
	return (ret);
}

static int _FAST		mm_get_steptime(  MMSQR_T *sqr, CONST char *s )
{
	int		step;

	/* Ĺʥƥåץ	*/
	{
		if ( *s == ':' )
		{
			MMV_T	mmv;
			++s;	/* å	*/

			if ( mm_getdigit( &mmv, &s ) || mmv < 0 )
			{
_STEP_ERR:
				sqr->errMsg = "Steptime must be identifier!";
			}
			step = mmv;
		} else if ( isdigit(*s) )
		{
			MMV_T	mmv;

			if ( mm_getdigit( &mmv, &s ) || mmv < 1 )
				goto _STEP_ERR;
			step = (sqr->division * 4) / mmv;
		} else
			goto _STEP_ERR;

		/* 	*/
		{
			int		st = 0;
			while ( *s )
			{
				if ( *s != '.' )
					break;
				{
					int		m;
					m = step;
					step = step + (step - st)/ 2;
					st = m;
				}
				++s;
			}
		}
	}
	return (step);
}
