/*************************************************************************
*	ȥ
*************************************************************************/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>

#include	"he386.h"
#include	"mmfile.h"
#include	"mml.h"
#include	"mmlib.h"
#include	"mmtrk.h"
#include	"mmptn.h"
#include	"mmctrl.h"
#include	"mmopt.h"
#include	"mmcmdtyp.h"
#include	"mmimport.h"
#include	"mmsg.h"
#include	"msgdat.h"

/*	config	------------------------------------------------------------*/
/*	#define	_DEBUG	*/
/*--------------------------------------------------------------------*/

static int _FAST	MMctr_title( MMSQR_T *sqr, CONST char *s );
static int _FAST	MMctr_copyright( MMSQR_T *sqr, CONST char *s );
static int _FAST	MMctr_division( MMSQR_T *sqr, CONST char *s );
static int _FAST	MMctr_tempo( MMSQR_T *sqr, CONST char *s );
static int _FAST	MMctr_signature( MMSQR_T *sqr, CONST char *s );
static int _FAST	MMctr_meas( MMSQR_T *sqr, CONST char *s );

static int _FAST	MMctr_track( MMSQR_T *sqr, CONST char *s );
static int _FAST	MMctr_pattern( MMSQR_T *sqr, CONST char *s );
static int _FAST	MMctr_mmlcmd( MMSQR_T *sqr, CONST char *s );
static int _FAST	MMctr_ccvbase( MMSQR_T *sqr, CONST char *s );
static int _FAST	MMctr_table( MMSQR_T *sqr, CONST char *s );

static CONST struct _mctrcmdtbl_t
{
	CONST char	   *name;
#if	defined(_MSC_VER) || defined(__WATCOMC__)
	int		(_FAST *func)(MMSQR_T *, CONST char *);
#else
	int _FAST     (*func)(MMSQR_T *, CONST char *);
#endif
} CtrCmdTbl[] =
{
	{ "ccvbase"  ,	MMctr_ccvbase  },
	{ "copyright",	MMctr_copyright},
	{ "division" ,	MMctr_division },
	{ "end"      ,	MMctr_end      },
	{ "import"   ,	MMctr_import   },
	{ "meas"     ,	MMctr_meas     },
	{ "mmlcmd"   ,	MMctr_mmlcmd   },
	{ "optset"   ,	MMctr_optset   },
	{ "pattern"  ,	MMctr_pattern  },
	{ "signature",	MMctr_signature},
	{ "table"    ,	MMctr_table    },
	{ "tempo"    ,	MMctr_tempo    },
	{ "title"    ,	MMctr_title    },
	{ "track"    ,	MMctr_track    },
};

static int		cmpFunc( const struct _mctrcmdtbl_t *c1, const struct _mctrcmdtbl_t *c2 )
{
	return (mm_strcmpi(c1->name,c2->name));
}

int _FAST	MMctr_line( MMSQR_T *sqr, CONST char *s )
{
	CONST char	*cmdname;

	if ( *s == '.' )
		++s;	/* skip '.'	*/
	while ( isspace(*s) )
		++s;
	if ( *s == '\0' )
	{	/* ȥԤθ	*/
		sqr->errMsg = "Control line error!!";
		return (ERR);
	}

	/* ޥ̾Ф	*/
	cmdname = s;
	for (;;)
	{
		if ( MM_ISSTRTERM(*s) )
			break;
		s++;
	}

	{	/* ̿ʬ	*/
		struct _mctrcmdtbl_t	key;
		struct _mctrcmdtbl_t	*cmd;

		key.name = cmdname;
#if	defined(_MSC_VER) || defined(__WATCOMC__)
		key.func = (int (_FAST *)(MMSQR_T *, CONST char *))0;
#else
		key.func = (int _FAST (*)(MMSQR_T *, CONST char *))0;
#endif
		cmd = bsearch( &key, CtrCmdTbl, sizeof(CtrCmdTbl)/sizeof(*CtrCmdTbl),
		  sizeof(*CtrCmdTbl), (int (*)(const void *, const void *))cmpFunc );
		if ( cmd )
		{
			return (cmd->func)(sqr,s);
		} else
		{
			sqr->errMsg = "Unknown control command!!";
			return (ERR);
		}
	}
}

/*************************************************************************
*	ȥ
*-------------------------------------------------------------------------
*	.title	<ȥ>
*************************************************************************/

static int _FAST	MMctr_title( MMSQR_T *sqr, CONST char *s )
{
	while ( isspace(*s) )
		++s;

	strncpy( sqr->title, s, _MMSQR_TITLE_LEN );
	sqr->title[_MMSQR_TITLE_LEN - 1] = '\0';
#ifdef	_DEBUG
	HE_printf(sqr,_HE_MSG_DE2,"--- MMctrl_title(\x22%s\x22)", sqr->title );
#endif

	return (NORMAL);
}

/*************************************************************************
*	ԡ饤Ȥ
*-------------------------------------------------------------------------
*	.copyright	<str>
*************************************************************************/

static int _FAST	MMctr_copyright( MMSQR_T *sqr, CONST char *s )
{
	SKIP_SPACE(s);
	strncpy( sqr->copyright, s, _MMSQR_COPYRIGHT_LEN );
	sqr->copyright[_MMSQR_COPYRIGHT_LEN - 1] = '\0';

	return (NORMAL);
}

/*************************************************************************
*	ʬǽ
*-------------------------------------------------------------------------
*	.division	<n>
*************************************************************************/

static int _FAST	MMctr_division( MMSQR_T *sqr, CONST char *s )
{
	int		division;

	if ( mm_getIntDigit( &division, &s ) || division < 24 || division > 9600 )
	{
		sqr->errMsg = "Division error!!";
		return (ERR);
	}
	sqr->division = division;
#ifdef	_DEBUG
	HE_printf(sqr,_HE_MSG_DE2,"--- MMctrl_division(%d)", sqr->division );
#endif

	return (NORMAL);
}

/*************************************************************************
*	ȥåץƥ
*-------------------------------------------------------------------------
*	.tempo	<n>
*************************************************************************/

static int _FAST	MMctr_tempo( MMSQR_T *sqr, CONST char *s )
{
	int		tempo;

	if ( mm_getIntDigit( &tempo, &s ) || tempo < MMTEMPO_MIN || tempo > MMTEMPO_MAX )
	{
		sqr->errMsg = "Tempo error!!";
		return (ERR);
	}
	sqr->top_tempo = 60000000 / tempo;
#ifdef	_DEBUG
	HE_printf(sqr,_HE_MSG_DE2,"--- MMctrl_tempo(%d)", sqr->top_tempo );
#endif

	return (NORMAL);
}

/*	Ѵơ֥	*/
static CONST char	SignTbl[32] =
{ /*  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 */
	  0, 1,-1, 2,-1,-1,-1, 3,-1,-1,-1,-1,-1,-1,-1, 4,

  /* 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32 */
	 -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 5,
};

/*************************************************************************
*	
*-------------------------------------------------------------------------
*	.signature	<n>/<d>
*************************************************************************/

static int _FAST	MMctr_signature( MMSQR_T *sqr, CONST char *s )
{
	int		sign_n, sign_d;

	if ( sqr->signSetFlag )
	{
_SIGN_ERR:
		sqr->errMsg = "Signature error!!";
		return (ERR);
	}
	if ( mm_getIntDigit( &sign_n, &s ) || sign_n < 1 || sign_n > 32 )
		goto _SIGN_ERR;
	if ( *s != '/' )
		goto _SIGN_ERR;
	++s;
	{
		if ( mm_getIntDigit( &sign_d, &s ) || sign_d < 1 || sign_d > 32 )
			goto _SIGN_ERR;
		if ( (sign_d = SignTbl[sign_d - 1]) < 0 )
			goto _SIGN_ERR;
	}
	{
		sqr->sign[0] = sqr->top_sign[0] = sign_n;
		sqr->sign[1] = sqr->top_sign[1] = sign_d;
	}
#ifdef	_DEBUG
	HE_printf(sqr,_HE_MSG_DE2,"--- MMctrl_signature(%d/%d)", sqr->sign[0], 1<<(sqr->sign[1]) );
#endif

	return (NORMAL);
}

static MMCLK_T _FAST	next_meas( MMSQR_T *sqr )
{
	int		i;
	MMCLK_T	clk;

	/* ƥȥåѡȤΥå·	*/
	clk = sqr->clk + 4 * sqr->division * sqr->sign[0] / (1<<sqr->sign[1]);
	for ( i = 0; i < _MMSQR_TRK_HASH_MAX; ++i )
	{
		MMTRK_T	*trk;
		trk = sqr->trkTbl[i];
		while ( trk )
		{
			MMtrk_setClock( sqr, trk, clk );
			if ( sqr->max_clk < trk->max_clk )
				sqr->max_clk = trk->max_clk;
			trk = trk->next;
		}
	}
	++(sqr->meas);
	{
		char	tmp[32];

		HE_printf(sqr, _HE_MSG_LIN, _msg_mmdec_compile_work_line[TypMsg],
		    clk, sqr->meas, sqr->sign[0], 1<<sqr->sign[1], GetSizeStr( sqr->workSize, tmp ) );
	}

	return (sqr->clk = clk);
}

/*************************************************************************
*	
*-------------------------------------------------------------------------
*	.meas	[<n>/<d>]
*************************************************************************/

static int _FAST	MMctr_meas( MMSQR_T *sqr, CONST char *s )
{
	MMCLK_T	clk;

	clk = next_meas(sqr);

	/* ξҤɤ߼	*/
	while ( isspace(*s) )
		++s;
	if ( *s )
	{
		int		sign_d, sign_n;

		if ( mm_getIntDigit( &sign_n, &s ) || sign_n < 1 || sign_n > 32 )
		{
_MEAS_ERR:
			sqr->errMsg = "\x22.meas\x22 error!!";
			return (ERR);
		}
		if ( *s != '/' )
			goto _MEAS_ERR;
		++s;
		{
			if ( mm_getIntDigit( &sign_d, &s ) || sign_d < 1 || sign_d > 32 )
				goto _MEAS_ERR;
			if ( (sign_d = SignTbl[sign_d - 1]) < 0 )
				goto _MEAS_ERR;
		}
		if ( sqr->sign[0] != sign_n || sqr->sign[1] != sign_d )
		{
			sqr->sign[0] = sign_n;
			sqr->sign[1] = sign_d;
			if ( mmtrk_set_signature( sqr->ctrlTrk, clk, sqr->sign ) )
				return (ERR);
		}
	}
#ifdef	_DEBUG
	HE_printf(sqr,_HE_MSG_DE2,"--- MMctrl_meas( %d/%d ) ", sqr->sign[0], 1<<(sqr->sign[1]) );
#endif
	return (NORMAL);
}

int _FAST	MMctr_end( MMSQR_T *sqr, CONST char *s )
{
	MMCLK_T	clk;

	do
	{
		clk = next_meas(sqr);
	} while ( clk < sqr->max_clk + 1 );

	if ( clk > 0 )
		--clk;
	return he386_genarate( sqr, clk );
}


/*************************************************************************
*	ȥå
*-------------------------------------------------------------------------
*	.track	<track no> port:<port> ch:<ch> macro:<name>
*************************************************************************/

static int _FAST	MMctr_track( MMSQR_T *sqr, CONST char *s )
{
	int			i;
	int			trkNo;
	int			port  = 0;
	int			ch    = 1;
	CONST char	*name = NULL;
	int			argc = 0;
	char		**argv = NULL;

	if ( mm_getIntDigit( &trkNo, &s ) || trkNo < 1 )
	{
TRK_ERR:
		if ( argc > 0 && argv != NULL )
			FS_argClear(argc,argv);
		sqr->errMsg = "Track setting error!";
		return (ERR);
	}
	if ( (argc = FS_argSet( s, &argv )) < 1 )
		goto TRK_ERR;

	for ( i = 0; i < argc; ++i )
	{
		CONST char	*p;
		p = argv[i];
		if ( mm_strncmpi( p, "port:", 5) == 0 )
		{
			p += 5;
			if ( (port = mm_getport( &p )) < 0 )
				goto TRK_ERR;
		} else if ( mm_strncmpi( p, "ch:", 3) == 0 )
		{
			p += 3;
			if ( (ch = mm_getmidich( &p )) < 0 )
				goto TRK_ERR;
		} else if ( mm_strncmpi( p, "macro:", 6) == 0 )
		{
			p += 6;
			name = p;
			if ( *name == '\0' )
				goto TRK_ERR;
		} else
			goto TRK_ERR;
	}
	if ( name == NULL )
		goto TRK_ERR;

	i = MMtrk_set( sqr, trkNo, name, port, ch );
	FS_argClear(argc,argv);
	return (i);
}

/*************************************************************************
*	ѥϿ
*-------------------------------------------------------------------------
*	.pattern	<pattern name>(<n>)	pattern
*************************************************************************/

static int _FAST	MMctr_pattern( MMSQR_T *sqr, CONST char *s )
{
	char		name[128];
	int			no;
	CONST char	*ptn;

	SKIP_SPACE(s);
	if ( mm_getname( name, 128, &s ) )
	{
_PTN_ERR:
		sqr->errMsg = "Pattern setting error!";
		return (ERR);
	}

	/* ѥ̾μФ	*/
	SKIP_SPACE(s);
	if ( *s != '(' )
		goto _PTN_ERR;
	++s;
	if ( mm_getIntDigit( &no, &s) || no < 1 )
		goto _PTN_ERR;
	SKIP_SPACE(s);
	if ( *s != ')' )
		goto _PTN_ERR;
	++s;

	/* ѥμФ	*/
	while ( isspace(*s) )
		++s;
	ptn = s;

	return MMptn_setPattern( sqr, name, no, ptn );
}


/*************************************************************************
*	MML ޥ
*-------------------------------------------------------------------------
*	.mmlcmd	<cha>	<command>
*************************************************************************/

static int _FAST	MMctr_mmlcmd( MMSQR_T *sqr, CONST char *s )
{
	int		n, typ;

	SKIP_SPACE(s);
	n = tolower(*s) - 'a';
	if ( n < 0 || n > 25 || sqr->mmlCmd[n] < 0 )
	{
_MMLCMD_ERR:
		sqr->errMsg = "MML command setting error!";
		return (ERR);
	}
	++s;
	/* ̤μФ	*/
	{
		SKIP_SPACE(s);
		if ( (typ = mm_get_mmlcmd( &s )) < 0 )
			goto _MMLCMD_ERR;
	}
	sqr->mmlCmd[n] = typ;

	return (NORMAL);
}


/*************************************************************************
*	ȥХ塼١
*-------------------------------------------------------------------------
*	.ccvbase	<trk no>	<cc>/<base>
*************************************************************************/

static int _FAST	MMctr_ccvbase( MMSQR_T *sqr, CONST char *s )
{
	MMTRK_T	   *trk;
	int			cc, base;

	if ( (trk = MMtrk_gettrk3( sqr, &s)) == NULL )
		return (-1);

	for (;;)
	{
		SKIP_SPACE(s);
		if ( *s == '\0' )
			break;

		if ( mm_getIntDigit(&cc,&s) || cc < 0 || cc > 127 )
			goto _CCV_ERR;
		if ( *s != '/' )
			goto _CCV_ERR;
		++s;	/* skip '/' */
		if ( mm_getIntDigit(&base,&s) || base < 0 || base > 127 )
			goto _CCV_ERR;

		trk->tbl.ccvBase[cc] = base;
	}

	return (NORMAL);

_CCV_ERR:
	sqr->errMsg = "Control value base setting error!";
	return (ERR);
}

/*************************************************************************
*	ơ֥
*-------------------------------------------------------------------------
*	.table	<trkno>	<typ>	<n1>-<n2>:<value>
*		<typ>	ch			MIDIͥ
*				note		
*				vel			
*				prog		
*				pitch		
*				cc			
*				ccv			
*************************************************************************/

static int _FAST	MMctr_table( MMSQR_T *sqr, CONST char *s )
{
	MMTRK_T	   *trk;
	int			typ = ERR;
	char		i, argc;
	char	  **argv;
	enum		_tbl_typ 
	{
		_TBL_CH,
		_TBL_NOTE,
		_TBL_VEL,
		_TBL_PITCH,
		_TBL_PROG,
		_TBL_CAF,
		_TBL_PAF,
		_TBL_CC,
		_TBL_CCV
	};
	static CONST struct
	{	CONST char	*name;
		int			typ;
	} tbl[] =
	{
		{	_ch_,		_TBL_CH		},
		{	_note_,		_TBL_NOTE	},
		{	_vel_,		_TBL_VEL	},
		{	_pitch_,	_TBL_PITCH	},
		{	_prog_,		_TBL_PROG	},
		{	_caf_,		_TBL_CAF	},
		{	_paf_,		_TBL_PAF	},
		{	_cc_,		_TBL_CC		},
	};

	if ( (trk = MMtrk_gettrk3( sqr, &s)) == NULL )
		return (-1);

	if ( (argc = FS_argSet( s, &argv)) < 2 )
		goto _TBL_ERR;

	{
		CONST char	*s = argv[0];

		if ( mm_strncmpi(s,"ccv:",4) == 0 )
		{
			int		cc;

			s += 4;
			cc = atoi(s);
			if ( cc < 0 || cc > 127 )
			{
_TYP_ERR:		sqr->errMsg = "Table type must be identifier!";
				return (ERR);
			}
			typ = _TBL_CCV  + cc;
		} else
		{
			for ( i = 0; i < sizeof(tbl)/sizeof(*tbl); ++i )
			{
				if ( strcmpi(s, tbl[i].name) == 0 )
				{
					typ = tbl[i].typ;
					break;
				}
			}
			if ( typ < 0 )
				goto _TYP_ERR;
		}
	}

	for ( i = 1; i < argc; ++i )
	{
		int			n, n1, n2, val;
		CONST char	*s;

		s = argv[i];
		if ( mm_getIntDigit(&n1,&s) )
		{
_PARA_ERR:	sqr->errMsg = "Table parameter error!";
			return (ERR);
		}
		if ( *s == '-' )
		{
			++s;	/* skip '-' */
			if ( mm_getIntDigit(&n2,&s) || n2 < n1 )
				goto _PARA_ERR;
		} else
			n2 = n1;
		if ( *s != ':' )
			goto _PARA_ERR;

		++s;	/* skip ':' */
		if ( mm_getIntDigit(&val,&s) )
		{
			goto _TBL_ERR;
		}
		for ( n = n1; n <= n2; ++n )
		{
			switch ( typ )
			{
				case	_TBL_CH:
					if ( n < 1 || n > 16 || val < 1 || val > 16 )
						goto _TBL_ERR;
					trk->tbl.ch[n - 1] = val - 1;
					break;

				case	_TBL_NOTE:
					if ( n < 0 || n > 127 || val < 0 || val > 127 )
						goto _TBL_ERR;
					trk->tbl.note[n] = val;
					break;

				case	_TBL_VEL:
					if ( n < 1 || n > 127 || val < 1 || val > 127 )
						goto _TBL_ERR;
					trk->tbl.vel[n] = val;
					break;

				case	_TBL_PITCH:
					if ( n < -8192 || n > 8191 || val < -8192 || val > 8191 )
						goto _TBL_ERR;
					if ( trk->extTbl.pitch == NULL )
					{	if ( mmtrk_set_pitchTable( trk ) )
							return (ERR);
					}
					trk->extTbl.pitch[n + 8192] = val + 8192;
					break;

				case	_TBL_PROG:
					if ( n < 1 || n > 128 || val < 1 || val > 128 )
						goto _TBL_ERR;
					trk->tbl.prog[n - 1] = val - 1;
					break;

				case	_TBL_CAF:
					if ( n < 0 || n > 127 || val < 0 || val > 127 )
						goto _TBL_ERR;
					trk->tbl.caf[n] = val;
					break;

				case	_TBL_PAF:
					if ( n < 0 || n > 127 || val < 0 || val > 127 )
						goto _TBL_ERR;
					trk->tbl.paf[n] = val;
					break;

				case	_TBL_CC:
					if ( n < 0 || n > 127 || val < 0 || val > 127 )
						goto _TBL_ERR;
					trk->tbl.cc[n] = val;
					break;

				default:
					if ( typ >= _TBL_CCV && typ <= _TBL_CCV+127 )
					{
						int		cc, base;

						cc = typ - _TBL_CCV;
						base = trk->tbl.ccvBase[cc];
						if ( n < -base || n > 127 - base
						  || val < -base || val > 127 - base )
							goto _TBL_ERR;
						if ( trk->extTbl.ccv[cc] == NULL )
						{
							if ( mmtrk_set_ccvTable( trk,cc ) )
								return (-1);
						}
						trk->extTbl.ccv[cc][n + base] = val + base;
					}
					break;
			}
		}
	}

	FS_argClear(argc,argv);

	return (NORMAL);

_TBL_ERR:
	sqr->errMsg = "Table setting error!";
	return (ERR);
}
