/*************************************************************************
*	ե륿
*************************************************************************/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>

#include	"he386.h"
#include	"midicc.h"
#include	"mml.h"
#include	"mmac.h"
#include	"mmlib.h"
#include	"mmpart.h"
#include	"mmtrk.h"
#include	"msgdat.h"

/*	config	------------------------------------------------------------*/
/*	#define	_DEBUG	*/
/*--------------------------------------------------------------------*/

#if	defined(CONFIG_UCWARN)
#	define	UCWARN(_expr)
#else
#	define	UCWARN(_expr)	_expr
#endif


#define	_FILTER_RATIO	(0)
#define	_FILTER_OFFSET	(1)
#define	_FILTER_MIN		(2)
#define	_FILTER_MAX		(3)
#define	_FILTER_BIT		(4)
#define	_FILTER_RND		(5)

static int _FAST	get_filter_para( CONST char *s, MMV_T *val1, MMV_T *val2 );

static int _FAST	get_filter_para( CONST char *s, MMV_T *val1, MMV_T *val2 )
{
	int		typ = ERR;
#if	defined(WIN32)
#	pragma	pack(push,4)
#endif
	static struct
	{
		CONST char	*name;
		int			typ;
	} tbl[] =
	{
		{ _MM_OFF_  , _FILTER_OFFSET },
		{ "ratio:"  , _FILTER_RATIO  },
		{ "min:"    , _FILTER_MIN    },
		{ "max:"    , _FILTER_MAX    },
		{ "bit:"    , _FILTER_BIT    },
		{ "rnd:"    , _FILTER_RND    },
	};
#if	defined(WIN32)
#	pragma	pack(pop)
#endif

	{
		int		i;

		for ( i = 0; i < sizeof(tbl)/sizeof(*tbl); ++i )
		{
			int		len;

			len = strlen(tbl[i].name);
			if ( mm_strncmpi( s, tbl[i].name, len) == 0 )
			{
				typ = tbl[i].typ;
				s += len;
				if ( mm_getdigit(val1,&s) )
					return (ERR);
				switch ( typ )
				{
					case _FILTER_RATIO:
						*val1 = (1024 * (*val1) + 99) / 100;
						break;
					case _FILTER_RND:
					{
						if ( *s != '/' )
							return (ERR);
						++s;
						if ( mm_getdigit(val2,&s) )
							return (ERR);
						break;
					}
				}
				break;
			}
		}
	}
	return (typ);
}

static int		filter_set_tempo(MMTRK_T *trk, int argc, char **argv);
static int		filter_set_delay(MMTRK_T *trk, int argc, char **argv);
static int		filter_set_note(MMTRK_T *trk, int argc, char **argv);
static int		filter_set_vel(MMTRK_T *trk, int argc, char **argv);
static int		filter_set_pitch(MMTRK_T *trk, int argc, char **argv);
static int		filter_set_prog(MMTRK_T *trk, int argc, char **argv);
static int		filter_set_caf(MMTRK_T *trk, int argc, char **argv);
static int		filter_set_paf(MMTRK_T *trk, int argc, char **argv);
static int		filter_set_cc(MMTRK_T *trk, int argc, char **argv);

/*************************************************************************
*	ե륿
*-------------------------------------------------------------------------
*	@FILTER (<cmd>, <par>, ...)
*		<cmd>	tempo
*				delay
*				note
*				vel
*				pitch
*				prog
*				cc:<cc>
*		<para>	off:<n>			եå
*				ratio:<n>		Ψ
*				min:<n>			Ǿ
*				max:<n>			
*				bit:<n>			ͭӥå
*				rnd:<n>/<n>		
*************************************************************************/
int		MMdec_trkFilter(MMSQR_T *sqr, MMPART_T *part, int argc, char **argv )
{
	MMTRK_T		*trk = part->trk;
#if	defined(WIN32)
#	pragma	pack(push,4)
#endif
	static struct
	{
		CONST char	*name;
		int			(*func)(MMTRK_T *,int, char **);
	}  tbl[] =
	{
		{ _tempo_  , filter_set_tempo},
		{ _delay_  , filter_set_delay},
		{ _note_   , filter_set_note },
		{ _vel_    , filter_set_vel  },
		{ _pitch_  , filter_set_pitch},
		{ _prog_   , filter_set_prog },
		{ _caf_    , filter_set_caf  },
		{ _paf_    , filter_set_paf  },
		{ _cc_     , filter_set_cc   },
	};
#if	defined(WIN32)
#	pragma	pack(pop)
#endif

	if ( argc < 2 )
	{
/*_FILTER_ERR:*/
		sqr->errMsg = "Filter error!!";
		return (ERR);
	}

	{
		int		i;
		char	*cmdname = argv[0];

		for ( i = 0; i < sizeof(tbl)/sizeof(*tbl); ++i )
		{
			int		len;

			len = strlen(tbl[i].name);
			if ( mm_strncmpi( cmdname, tbl[i].name, len) == 0 )
			{
				if ( (tbl[i].func)(trk,argc,argv) )
					return (ERR);
				break;
			}
		}
	}

	return (NORMAL);
}

static int		filter_set_tempo(MMTRK_T *trk, int argc, char **argv)
{
	int				i;
	MMFILTER32_T	*filter = &(trk->filter.tempo);

	for ( i = 1; i < argc; ++i )
	{
		MMV_T	val1, val2;
		switch( get_filter_para( argv[i], &val1, &val2) )
		{
			case _FILTER_RATIO:		filter->ratio   =        val1;	break;
			case _FILTER_OFFSET:	filter->offset  =        val1;	break;
			case _FILTER_MIN:		filter->min     =        val1;	break;
			case _FILTER_MAX:		filter->max     =        val1;	break;
			case _FILTER_BIT:		filter->bit     = (UCHAR)val1;	break;
			case _FILTER_RND:
				if ( filter->att == 0 )
					return (ERR);
				filter->rnd = ( val1 == val2 ) ? 0 : 1;
				filter->rnd_min = val1;
				filter->rnd_max = val2;
				break;
			default:
				return (ERR);
		}
	}
	if ( (UCWARN(filter->bit < 0 ||) filter->bit > filter->max_bit)
	   || filter->min > filter->max
	   || filter->min < MMTEMPO_MIN || filter->max > MMTEMPO_MAX
	   || (filter->rnd &&
	       filter->rnd_min < MMTEMPO_MIN || filter->rnd_max > MMTEMPO_MAX
	      )
	)
		return (ERR);
	
	return (NORMAL);
}

static int		filter_set_delay(MMTRK_T *trk, int argc, char **argv)
{
	int				i;
	MMFILTER16_T	*filter = &(trk->filter.delay);

	for ( i = 1; i < argc; ++i )
	{
		MMV_T	val1, val2;
		switch( get_filter_para( argv[i], &val1, &val2) )
		{
			case _FILTER_RATIO:		filter->ratio   =        val1;	break;
			case _FILTER_OFFSET:	filter->offset  = (SHORT)val1;	break;
			case _FILTER_MIN:		filter->min     = (SHORT)val1;	break;
			case _FILTER_MAX:		filter->max     = (SHORT)val1;	break;
			case _FILTER_BIT:		filter->bit     = (UCHAR)val1;	break;
			case _FILTER_RND:
				if ( filter->att == 0 )
					return (ERR);
				filter->rnd = ( val1 == val2 ) ? 0 : 1;
				filter->rnd_min = (SHORT)val1;
				filter->rnd_max = (SHORT)val2;
				break;
			default:
				return (ERR);
		}
	}
	if ( (UCWARN(filter->bit < 0 ||) filter->bit > filter->max_bit)
	   || filter->min > filter->max
UCWARN(|| filter->min < -32767 || filter->max > 32767
	   || (filter->rnd &&
	       filter->rnd_min < -32767 || filter->rnd_max > 32767
	      ))
	)
		return (ERR);
	
	return (NORMAL);
}

static int		filter_set_note(MMTRK_T *trk, int argc, char **argv)
{
	int				i;
	MMFILTER08_T	*filter = &(trk->filter.note);

	for ( i = 1; i < argc; ++i )
	{
		MMV_T	val1, val2;
		switch( get_filter_para( argv[i], &val1, &val2) )
		{
			case _FILTER_RATIO:		filter->ratio   =        val1;	break;
			case _FILTER_OFFSET:	filter->offset  = (char) val1;	break;
			case _FILTER_MIN:		filter->min     = (char) val1;	break;
			case _FILTER_MAX:		filter->max     = (char) val1;	break;
			case _FILTER_BIT:		filter->bit     = (UCHAR)val1;	break;
			case _FILTER_RND:
				if ( filter->att == 0 )
					return (ERR);
				filter->rnd = ( val1 == val2 ) ? 0 : 1;
				filter->rnd_min = (char) val1;
				filter->rnd_max = (char) val2;
				break;
			default:
				return (ERR);
		}
	}
	if ( (UCWARN(filter->bit < 0 ||) filter->bit > filter->max_bit)
	   || filter->min > filter->max
UCWARN(|| filter->min < -127 || filter->max > 127
	   || (filter->rnd &&
	       filter->rnd_min < -127 || filter->rnd_max > 127
	      ))
	)
		return (ERR);
	
	return (NORMAL);
}

static int		filter_set_cc(MMTRK_T *trk, int argc, char **argv)
{
	int				i;
	MMFILTER08_T	*filter;

	{
		int			cc;
		CONST char	*s = argv[0] ;

		s += 2;	/* skip 'CC' */
		if ( *s != ':' )
			return (ERR);
		++s;
		if ( mm_getIntDigit( &cc, &s) || cc < 0 || cc > 127 )
			return (ERR);
		filter = &(trk->filter.ccv[cc]);
	}

	for ( i = 1; i < argc; ++i )
	{
		MMV_T	val1, val2;
		switch( get_filter_para( argv[i], &val1, &val2) )
		{
			case _FILTER_RATIO:		filter->ratio   =        val1;	break;
			case _FILTER_OFFSET:	filter->offset  = (char) val1;	break;
			case _FILTER_MIN:		filter->min     = (char) val1;	break;
			case _FILTER_MAX:		filter->max     = (char) val1;	break;
			case _FILTER_BIT:		filter->bit     = (UCHAR)val1;	break;
			case _FILTER_RND:
				if ( filter->att == 0 )
					return (ERR);
				filter->rnd = ( val1 == val2 ) ? 0 : 1;
				filter->rnd_min = (char)val1;
				filter->rnd_max = (char)val2;
				break;
			default:
				return (ERR);
		}
	}
	if ( (UCWARN(filter->bit < 0 ||) filter->bit > filter->max_bit)
	   || filter->min > filter->max
UCWARN(|| filter->min < -127 || filter->max > 127
	   || (filter->rnd &&
	       filter->rnd_min < -127 || filter->rnd_max > 127
	      ))
	)
		return (ERR);
	
	return (NORMAL);
}

static int		filter_set_prog(MMTRK_T *trk, int argc, char **argv)
{
	int				i;
	MMFILTER08_T	*filter = &(trk->filter.prog);

	for ( i = 1; i < argc; ++i )
	{
		MMV_T	val1, val2;
		switch( get_filter_para( argv[i], &val1, &val2) )
		{
			case _FILTER_RATIO:		filter->ratio   =        val1;	break;
			case _FILTER_OFFSET:	filter->offset  = (char) val1;	break;
			case _FILTER_MIN:		filter->min     = (char) val1;	break;
			case _FILTER_MAX:		filter->max     = (char) val1;	break;
			case _FILTER_BIT:		filter->bit     = (UCHAR)val1;	break;
			case _FILTER_RND:
				if ( filter->att == 0 )
					return (ERR);
				filter->rnd = ( val1 == val2 ) ? 0 : 1;
				filter->rnd_min = (char)val1;
				filter->rnd_max = (char)val2;
				break;
			default:
				return (ERR);
		}
	}
	if ( (UCWARN(filter->bit < 0 ||) filter->bit > filter->max_bit)
	   || filter->min > filter->max
UCWARN(|| filter->min < -127 || filter->max > 127
	   || (filter->rnd &&
	       filter->rnd_min < -127 || filter->rnd_max > 127
	      ))
	)
		return (ERR);
	
	return (NORMAL);
}

static int		filter_set_caf(MMTRK_T *trk, int argc, char **argv)
{
	int				i;
	MMFILTER08_T	*filter = &(trk->filter.caf);

	for ( i = 1; i < argc; ++i )
	{
		MMV_T	val1, val2;
		switch( get_filter_para( argv[i], &val1, &val2) )
		{
			case _FILTER_RATIO:		filter->ratio   =        val1;	break;
			case _FILTER_OFFSET:	filter->offset  = (char) val1;	break;
			case _FILTER_MIN:		filter->min     = (char) val1;	break;
			case _FILTER_MAX:		filter->max     = (char) val1;	break;
			case _FILTER_BIT:		filter->bit     = (UCHAR)val1;	break;
			case _FILTER_RND:
				if ( filter->att == 0 )
					return (ERR);
				filter->rnd = ( val1 == val2 ) ? 0 : 1;
				filter->rnd_min = (char)val1;
				filter->rnd_max = (char)val2;
				break;
			default:
				return (ERR);
		}
	}
	if ( (UCWARN(filter->bit < 0 ||) filter->bit > filter->max_bit)
	   || filter->min > filter->max
UCWARN(|| filter->min < -127 || filter->max > 127
	   || (filter->rnd &&
	       filter->rnd_min < -127 || filter->rnd_max > 127
	      ))
	)
		return (ERR);
	
	return (NORMAL);
}

static int		filter_set_paf(MMTRK_T *trk, int argc, char **argv)
{
	int				i;
	MMFILTER08_T	*filter = &(trk->filter.paf);

	for ( i = 1; i < argc; ++i )
	{
		MMV_T	val1, val2;
		switch( get_filter_para( argv[i], &val1, &val2) )
		{
			case _FILTER_RATIO:		filter->ratio   =        val1;	break;
			case _FILTER_OFFSET:	filter->offset  = (char) val1;	break;
			case _FILTER_MIN:		filter->min     = (char) val1;	break;
			case _FILTER_MAX:		filter->max     = (char) val1;	break;
			case _FILTER_BIT:		filter->bit     = (UCHAR)val1;	break;
			case _FILTER_RND:
				if ( filter->att == 0 )
					return (ERR);
				filter->rnd = ( val1 == val2 ) ? 0 : 1;
				filter->rnd_min = (char)val1;
				filter->rnd_max = (char)val2;
				break;
			default:
				return (ERR);
		}
	}
	if ( (UCWARN(filter->bit < 0 ||) filter->bit > filter->max_bit)
	   || filter->min > filter->max
UCWARN(|| filter->min < -127 || filter->max > 127
	   || (filter->rnd &&
	       filter->rnd_min < -127 || filter->rnd_max > 127
	      ))
	)
		return (ERR);
	
	return (NORMAL);
}

static int		filter_set_pitch(MMTRK_T *trk, int argc, char **argv)
{
	int				i;
	MMFILTER16_T	*filter = &(trk->filter.pitch);

	for ( i = 1; i < argc; ++i )
	{
		MMV_T	val1, val2;
		switch( get_filter_para( argv[i], &val1, &val2) )
		{
			case _FILTER_RATIO:		filter->ratio   =        val1;	break;
			case _FILTER_OFFSET:	filter->offset  = (SHORT)val1;	break;
			case _FILTER_MIN:		filter->min     = (SHORT)val1;	break;
			case _FILTER_MAX:		filter->max     = (SHORT)val1;	break;
			case _FILTER_BIT:		filter->bit     = (UCHAR)val1;	break;
			case _FILTER_RND:
				if ( filter->att == 0 )
					return (ERR);
				filter->rnd = ( val1 == val2 ) ? 0 : 1;
				filter->rnd_min = (SHORT)val1;
				filter->rnd_max = (SHORT)val2;
				break;
			default:
				return (ERR);
		}
	}
	if ( (UCWARN(filter->bit < 0 ||) filter->bit > filter->max_bit)
	   || filter->min > filter->max
	   || filter->min < -16383 || filter->max > 16383
	   || (filter->rnd &&
	       filter->rnd_min < -16383 || filter->rnd_max > 16383
	      )
	)
		return (ERR);
	
	return (NORMAL);
}

static int		filter_set_vel(MMTRK_T *trk, int argc, char **argv)
{
	int				i;
	MMFILTER08_T	*filter = &(trk->filter.vel);

	for ( i = 1; i < argc; ++i )
	{
		MMV_T	val1, val2;
		switch( get_filter_para( argv[i], &val1, &val2) )
		{
			case _FILTER_RATIO:		filter->ratio   =        val1;	break;
			case _FILTER_OFFSET:	filter->offset  = (char) val1;	break;
			case _FILTER_MIN:		filter->min     = (char) val1;	break;
			case _FILTER_MAX:		filter->max     = (char) val1;	break;
			case _FILTER_BIT:		filter->bit     = (UCHAR)val1;	break;
			case _FILTER_RND:
				if ( filter->att == 0 )
					return (ERR);
				filter->rnd = ( val1 == val2 ) ? 0 : 1;
				filter->rnd_min = (char)val1;
				filter->rnd_max = (char)val2;
				break;
			default:
				return (ERR);
		}
	}
	if ( (UCWARN(filter->bit < 0 ||) filter->bit > filter->max_bit)
	   || filter->min > filter->max
UCWARN(|| filter->min < -127 || filter->max > 127
	   || (filter->rnd &&
	       filter->rnd_min < -127 || filter->rnd_max > 127
	      ))
	)
		return (ERR);
	
	return (NORMAL);
}
