/*************************************************************************
*		饤֥
*************************************************************************/

#include	<stdio.h>
#include	<stdlib.h>
#include	<stdarg.h>
#include	<ctype.h>
#include	<string.h>
#include	<time.h>

#ifdef	__HIGHC__
#	include	<spdefs.h>
#	include	<spfunc.h>
#	include	<fslib.h>
#	include	<guidefs.h>
#endif

#include	"he386.h"
#include	"msgdat.h"
#include	"mml.h"
#include	"mmcmdtyp.h"
#include	"mmlib.h"
#include	"mmpp.h"
#include	"mmac.h"
#if	defined(__HE32__)
#	include	"mmfile.h"
#endif
#include	"search.h"

#ifdef	__HIGHC__
#	pragma	On(Align_labels);
#endif

/*	config	------------------------------------------------------------*/
/*	#define	_DEBUG	*/
/*--------------------------------------------------------------------*/

void	MM_error( MMSQR_T *sqr )
{
	he386_error();
}

void	MM_setErrMsg( MMSQR_T *sqr, CONST char *form, ... )
{
	va_list		arg;

	va_start( arg, form );
	vsprintf( sqr->errMsgBuf, form, arg );
	va_end(arg);
	sqr->errMsg = sqr->errMsgBuf;
}

void	*MM_fopen( MMSQR_T *sqr, CONST char *fn, CONST char *md )
{
	FILE	*fp;

	if ( (fp = SP_FOPEN(fn,md)) == NULL )
	{
		char	tmp[256];
		int		len;

		FS_getDnamPath(tmp,sqr->inc[0].fn);
		if ( (len = strlen(tmp)) > 0 )
		{
			if ( !FS_ISDELIM(tmp[len-1]) )
				tmp[len++] = FsPathDelim;
			strcpy( &tmp[len], fn );
			fp = SP_FOPEN(tmp,md);
		}
	}
	return (fp);
}

int _FAST	mm_strcmp( CONST char *s1, CONST char *s2 )
{
	int		i, a1, a2;

	for (;;)
	{
		a1 = MM_ISSTRTERM(*s1);
		a2 = MM_ISSTRTERM(*s2);
		if ( a1 && a2 )
			return (0);
		i = (*s1) - (*s2);
		if ( i )
			return (i);
		++s1;
		++s2;
	}
}

int _FAST	mm_strcmpi( CONST char *s1, CONST char *s2 )
{
	int		i, a1, a2;

	for (;;)
	{
		a1 = MM_ISSTRTERM(*s1);
		a2 = MM_ISSTRTERM(*s2);
		if ( a1 && a2 )
			return (0);
		i = tolower(*s1) - tolower(*s2);
		if ( i )
			return (i);
		++s1;
		++s2;
	}
}

int _FAST	mm_strncmpi( CONST char *s1, CONST char *s2, size_t len )
{
	int		i, a1, a2;

	while ( len > 0 )
	{
		a1 = MM_ISSTRTERM(*s1);
		a2 = MM_ISSTRTERM(*s2);
		if ( a1 && a2 )
			return (0);
		i = tolower(*s1) - tolower(*s2);
		if ( i )
			return (i);
		++s1;
		++s2;
		--len;
	}
	return (0);
}


MMV_T _FAST	mm_getXdigit( int ch )
{
	if ( isdigit(ch) )
		return (ch - '0');
	if ( ch >= 'A' && ch <= 'F' )
		return (10 + ch - 'A');
	if ( ch >= 'a' && ch <= 'f' )
		return (10 + ch - 'a');
	return (-1);
}

int _FAST	mm_getdigit( MMV_T *ans, CONST char **s )
{
	int		sign = 1;

	while ( isspace(*(*s)) )
		++(*s);
	if ( *(*s) == '-' )
	{
		++(*s);
		sign = -1;
	} else if ( *(*s) == '+' )
		++(*s);
	*ans = 0;
	if ( !isdigit(*(*s)) )
		return (-1);
	while ( isdigit(*(*s)) )
	{
		*ans = (*ans) * 10 + *(*s) - '0';
		++(*s);
	}
	if ( sign < 0 )
		*ans = -(*ans);

	return (0);
}

int _FAST	mm_getIntDigit( int *ans, CONST char **s )
{
	int		sign = 1;

	while ( isspace(*(*s)) )
		++(*s);
	if ( *(*s) == '-' )
	{
		++(*s);
		sign = -1;
	} else if ( *(*s) == '+' )
		++(*s);
	*ans = 0;
	if ( !isdigit(*(*s)) )
		return (-1);
	while ( isdigit(*(*s)) )
	{
		*ans = (*ans) * 10 + *(*s) - '0';
		++(*s);
	}
	if ( sign < 0 )
		*ans = -(*ans);

	return (0);
}

int _FAST	mm_getname( char *buf, size_t bufsiz, CONST char **s )
{
	size_t	len = 0;

	SKIP_SPACE(*s);
	if ( (!isalpha(*(*s)) || *(*s) == '_') )
		return (-1);
	while ( *(*s) )
	{
		if ( !(isalpha(*(*s)) || *(*s) == '_' || isdigit(*(*s))) )
			break;
		buf[len++] = *(*s);
		++(*s);
		if ( len >= bufsiz )
			return (-1);
	}
	buf[len] = '\0';

	return (0);
}

MMV_T	MM_setValue( MMVALUE_T *val, int mode, ... )
{
	va_list		ap;

	va_start(ap,mode);
	switch( val->mode = mode )
	{
		case	_MMVAL_ABS:			/* л	*/
			val->value = va_arg(ap,int);
			break;

		case	_MMVAL_RELATIVE:	/* Ѳ	*/
		{
			int		relative = va_arg(ap,int);
			val->value += relative;
			break;
		}
		case	_MMVAL_RATIO:
		{
			int		ratio = va_arg(ap,int);
			val->value = val->value * ratio / _MMVAL_RATIO_BASE;
			break;
		}
		case	_MMVAL_RANDOM:
			val->rnd_min = va_arg(ap,int);
			val->rnd_max = va_arg(ap,int);
			{
				int		rnd_h;

				rnd_h = val->rnd_max -  val->rnd_min;
				if ( rnd_h == 0 )
					val->value = val->rnd_min;
				else if ( rnd_h > 0 )
					val->value = val->rnd_min + (mm_rand() % (rnd_h+1));
			}
			break;

		case	_MMVAL_RELATIVE | _MMVAL_RATIO:	/* ΨѲ	*/
		{
			int		ratio = va_arg(ap,int);
			val->value += val->value * ratio / _MMVAL_RATIO_BASE;
			break;
		}

		case	_MMVAL_RELATIVE | _MMVAL_RANDOM:	/* Х	*/
			val->rnd_min = va_arg(ap,int);
			val->rnd_max = va_arg(ap,int);
			{
				int		rnd_h;

				rnd_h = val->rnd_max -  val->rnd_min;
				if ( rnd_h == 0 )
					val->value += val->rnd_min;
				else if ( rnd_h > 0 )
					val->value += ( val->rnd_min + (mm_rand() % (rnd_h+1)) );
			}
			break;

		case	_MMVAL_RATIO | _MMVAL_RANDOM:	/* Ψ	*/
			val->rnd_min = va_arg(ap,int);
			val->rnd_max = va_arg(ap,int);
			{
				int		rnd_h;

				rnd_h = val->rnd_max -  val->rnd_min;
				if ( rnd_h == 0 )
					val->value = val->value * val->rnd_min / _MMVAL_RATIO_BASE;
				else if ( rnd_h > 0 )
					val->value = val->value * ( val->rnd_min + (mm_rand() % (rnd_h+1)) ) / _MMVAL_RATIO_BASE;
			}
			break;

		case	_MMVAL_RELATIVE | _MMVAL_RATIO | _MMVAL_RANDOM:	/* Ψ	*/
			val->rnd_min = va_arg(ap,int);
			val->rnd_max = va_arg(ap,int);
			{
				int		rnd_h;

				rnd_h = val->rnd_max -  val->rnd_min;
				if ( rnd_h == 0 )
					val->value += val->value * val->rnd_min / _MMVAL_RATIO_BASE;
				else if ( rnd_h > 0 )
					val->value += val->value * ( val->rnd_min + (mm_rand() % (rnd_h+1)) ) / _MMVAL_RATIO_BASE;
			}
			break;
	}
	va_end(ap);

	return (val->value);
}

/*************************************************************************
*	ĥͻɤ߼
*************************************************************************/

int _FAST	mm_getValue( MMVALUE_T *val, CONST char **s )
{
	int		mode = _MMVAL_ABS;
	MMV_T	val1, val2;

	SKIP_SPACE(*s);

	while ( !(isdigit(*(*s)) || *(*s) == '+' || *(*s) == '-') )
	{
		if ( *(*s) == '\'' )
		{	/* л	*/
			mode &= ~(_MMVAL_ABS);
			mode |= _MMVAL_RELATIVE;
			++(*s);
		} else
		{	if ( *(*s) == '*' )
			{	/* 	*/
				mode &= ~(_MMVAL_ABS);
				mode |= _MMVAL_RANDOM;
				++(*s);
				if ( *(*s) == '\'' )
				{	/* л	*/
					mode |= _MMVAL_RELATIVE;
					++(*s);
				}
			} else
				break;
		}
	}
	if ( mm_getdigit( &val1, s) )
		return (-1);

	SKIP_SPACE(*s);
	if ( mode & _MMVAL_RANDOM )
	{	/* 	*/
		if ( *(*s) != '.' )	return (-1);
		++(*s);
		if ( *(*s) != '.' )	return (-1);
		++(*s);
		SKIP_SPACE(*s);
		if ( mm_getdigit( &val2, s) )
			return (-1);
		SKIP_SPACE(*s);
		if ( val1 > val2 )	/* Append ver.3.01f #02 1993/11/23 */
		{
			MMV_T	m;

			m = val1;
			val1 = val2;
			val2 = m;
		}
	} else
		val2 = 0;

	if ( *(*s) == '%' )
	{	/* Ψ	*/
		mode &= ~(_MMVAL_ABS);
		mode |= _MMVAL_RATIO;
		++(*s);
		val1 = _MMVAL_RATIO_BASE * val1 / 100;
		val2 = _MMVAL_RATIO_BASE * val2 / 100;
	}

	if ( mode & _MMVAL_RANDOM )
		MM_setValue( val, mode, val1, val2 );
	else
		MM_setValue( val, mode, val1 );

	return (0);
}


/*************************************************************************
*	ĥͻɤ߼
*	СΨѲͤФƤ
*************************************************************************/

int _FAST	mm_getValue2( MMVALUE_T *val, CONST char **s )
{
	int		mode = _MMVAL_ABS;
	MMV_T	val1, val2;

	SKIP_SPACE(*s);

	while ( !(isdigit(*(*s)) || *(*s) == '+' || *(*s) == '-') )
	{
		if ( *(*s) == '\'' )
		{	/* л	*/
			mode &= ~(_MMVAL_ABS);
			mode |= _MMVAL_RELATIVE;
			++(*s);
		} else
		{	if ( *(*s) == '*' )
			{	/* 	*/
				mode &= ~(_MMVAL_ABS);
				mode |= _MMVAL_RANDOM;
				++(*s);
				if ( *(*s) == '\'' )
				{	/* л	*/
					mode |= _MMVAL_RELATIVE;
					++(*s);
				}
			} else
				break;
		}
	}
	if ( mm_getdigit( &val1, s) )
		return (-1);

	SKIP_SPACE(*s);
	if ( mode & _MMVAL_RANDOM )
	{	/* 	*/
		if ( *(*s) != '.' )	return (-1);
		++(*s);
		if ( *(*s) != '.' )	return (-1);
		++(*s);
		SKIP_SPACE(*s);
		if ( mm_getdigit( &val2, s) )
			return (-1);
		SKIP_SPACE(*s);
	} else
		val2 = 0;

	if ( *(*s) == '%' )
	{	/* Ψ	*/
		mode &= ~(_MMVAL_ABS);
		mode |= _MMVAL_RATIO;
		++(*s);
		val1 = _MMVAL_RATIO_BASE * val1 / 100;
		val2 = _MMVAL_RATIO_BASE * val2 / 100;
	}

	if ( (mode & _MMVAL_RELATIVE) || (mode & _MMVAL_RATIO) )
	{
		if ( mode & _MMVAL_RANDOM )
		{
			val->rnd_min = val1;
			val->rnd_max = val2;
		} else
			val->value   = val1;
		val->mode = mode;
	} else
	{
		if ( mode & _MMVAL_RANDOM )
			MM_setValue( val, mode, val1, val2 );
		else
			MM_setValue( val, mode, val1 );
	}
	return (0);
}

/*************************************************************************
*	̡'(',')'Υѥ᡼
*-------------------------------------------------------------------------
*	(*s) '(' ؤ֤ǸƤӽФ
*	(*s) ')'μΥݥ󥿤򼨤
*************************************************************************/

int _FAST	mm_get_arg( CONST char **sp, char **argv[] )
{
	int			len = 0;
	CONST char	*s = *sp;

	if ( *s != '(' )
		return (0);

	{
		int		kl = 0;
		CONST char	*p;

		p = ++s;	/* skip '(' */
		while ( *p )
		{
			if ( *p == '(' )
			{
				++kl;
			} else if ( *p == ')' )
			{
				if ( kl == 0 )
				{	++p;
					*sp = p;	/* ݥ󥿹	*/
					break;
				}
				--kl;
			} else if ( *p == '\0' )
			{
				*sp = p;					/****** ERROR *****/
				return (0);
			}
			++p;
			++len;
		}
	}

	{
		int			i;
		char		*buff;
		char		top_flg = TRUE;
		char		cma_flg = FALSE;	/* ',' flag	*/
		int			argc = 0;			/* ñο */
		char		*strBuf;

		if ( len == 0 )
			return (0);
		if ( (strBuf = malloc(len + 1)) == NULL )
			return (0);
		buff = strBuf;

		for(;;)
		{
			int		ch;

			if ( len )
			{
				--len;
				ch = *(unsigned char *)s++;
			} else
				ch = 0;
			if ( top_flg )			/* ñƬ */
			{
				if ( ch == '\0' )
				{	if ( cma_flg )	/* ľ(,) ʤ */
						goto SKIP_0;
					goto SKIP;
				}
				if ( ch <= ' ' )		/* Skip Sapce */
					goto SKIP;
				if ( ch == ',' )
				{
					cma_flg =TRUE;
					goto SKIP_0;
				}
			}

			/* ñ */
			top_flg = FALSE;
			cma_flg = FALSE;
			if ( ch == '\0' )
				goto SKIP_0;

			if ( ch == ',' )
			{
				cma_flg = TRUE;
				top_flg = TRUE;
				goto SKIP_0;
			} else if ( ch == '"' )
			{
/*DQMARKS:*/	/*  "  */
				for(;;)
				{
					if ( len )
					{
						--len;
						ch = *s++;
					} else
						ch = 0;
					if ( ch == '\0' || ch == '"' )
						goto SKIP;		/* End of Job */
					*buff++ = ch;
				}
			}
			*buff++ = ch;
			goto SKIP;
  
SKIP_0:		++argc;
			*buff++ ='\0';
SKIP:		if ( ch == '\0' )
				break;
		} /* LOOP */

		{
			char		**strPtr;
			int			j, k;

			if ( argc == 0 || (strPtr = malloc( sizeof(char *) * (argc+1) )) == NULL )
			{
				free( strBuf );
				return (0);
			}

			for ( i = j = k = 0; (size_t)k <= (size_t)(buff - strBuf); ++k )
			{
				if ( strBuf[k] == '\0' )
				{
					if ( (strPtr[i] = malloc( strlen(&strBuf[j]) + 1) ) == NULL )
					{
						while ( --i >= 0 )
							free( strPtr[i] );
						free( strPtr );
						free( strBuf );
						return (0);
					}
					strcpy( strPtr[i], &strBuf[j] );
					++i;
					if ( i >= argc )
					{
						strPtr[argc] = NULL;
						break;
					}
					j = k + 1;
				}
			}
			free( strBuf );
			*argv = strPtr;
		}
		return	(argc);
	}
}

void _FAST	mm_clear_arg(int argc,char **argv)
{
	if ( argc > 0 && argv )
	{
		while ( --argc >= 0 )
			free( argv[argc] );
		free( argv );
	}
}

char * _FAST	mm_getTimeStr( time_t tm )
{
	struct tm	*tp;
	static char	buf[64];

	tm = time(NULL);
	tp = localtime(&tm);
	sprintf( buf, "%4d/%02d/%02d %02d:%02d:%02d",
	    tp->tm_year +  1900,
	    tp->tm_mon + 1,
	    tp->tm_mday,
	    tp->tm_hour,
	    tp->tm_min,
	    tp->tm_sec );
	return (buf);
}

int _FAST	mm_chkOnOff( const char *s )
{
	if ( mm_strcmpi(s, _on_) == 0 )
		return (1);
	else if ( mm_strcmpi(s, _off_) == 0 )
		return (0);
	return (-1);
}

/*************************************************************************
*		ݡȤɤ߼
*		<18> or <AH>
*		return  0  7
*************************************************************************/

int _FAST	mm_getport( CONST char **p )
{
	int			port = -1;

	SKIP_SPACE(*p);
	if ( isalpha(*(*p)) )
	{
		port = toupper(*(*p)) - 'A';
		if ( port >= 0 && port <= 7 )
			return (port);
	} else if ( isdigit(*(*p)) )
	{
		if ( mm_getIntDigit( &port, p ) == NORMAL )
		{
			if ( port >= 1 && port <= 8 )
				return (port - 1);
		}
	}
	return (-1);
}

/*************************************************************************
*		MIDI CHɤ߼
*		<116>
*		return  0  15
*************************************************************************/

int _FAST	mm_getmidich( CONST char **p )
{
	int		ch;

	SKIP_SPACE(*p);
	if ( mm_getIntDigit( &ch, p ) == NORMAL )
	{
		if ( ch >= 1 && ch <= 16 )
			return (ch - 1);
	}
	return (-1);
}


/*************************************************************************
*	ɤ߼
*		<0127> or  <A><octave>
*		return  0  15
*************************************************************************/

int _FAST	mm_getnote( CONST char **p )
{
	int		note;

	SKIP_SPACE(*p);
	if ( isdigit(*(*p)) )
	{
		if ( mm_getIntDigit( &note, p ) || note < 0 || note > 127 )
			return (-1);
	} else
	{
		int		key;
		{
			int		ch = tolower(*(*p));

			if ( ch < 'a' || ch > 'g' )
				return (ERR);
			++(*p);
			{
				static	int	tbl[7] = {9,11,0,2,4,5,7};

				key = tbl[ch - 'a'];
			}
		}
		{
			int		octave;
			if ( mm_getIntDigit( &octave, p )
			    || octave < MMOCTAVE_MIN || octave > MMOCTAVE_MAX )
				return (-1);
			note = key + 12 + (octave * 12);
		}
	}
	return (note);
}

/*************************************************************************
*	<n1>-<n2>
*************************************************************************/

int _FAST	mm_getnote_range( int *n1, int *n2, CONST char **s )
{
	if ( (*n1 = mm_getnote(s)) < 0 )
		return (ERR);
	if ( *(*s) != '-' )
	{
		*n2 = *n1;
		return (NORMAL);
	}
	++(*s);
	if ( (*n2 = mm_getnote(s)) < 0 || n2 < n1 )
		return (ERR);
	return (NORMAL);
}

/*************************************************************************
*	<p1>-<p2>
*************************************************************************/

int _FAST	mm_getprog_range( int *p1, int *p2, CONST char **s )
{
	int		prog;

	if ( mm_getIntDigit(&prog,s) || prog < 1 || prog > 128 )
		return (ERR);
	*p1 = prog - 1;
	if ( *(*s) != '-' )
	{	*p2 = *p1;
		return (NORMAL);
	}
	++(*s);	/* skip '-' */
	if ( mm_getIntDigit(&prog,s) || prog < 1 || prog > 128 )
		return (ERR);
	*p2 = prog - 1;
	if ( p2 < p1 )
		return (ERR);
	return (NORMAL);
}

/*************************************************************************
*	MML ޥ
*************************************************************************/

static int		compfunc_get_mmlcmd( const SEA_NO *key1, const SEA_NO *key2);

int _FAST	mm_get_mmlcmd( CONST char **s )
{
	int		typ = ERR;

	if ( mm_strncmpi(*s,"ccv:",4) == 0 )
	{	/* ȥ	*/
		int		cc;

		(*s) += 4;
		if ( mm_getIntDigit(&cc,s) || cc < 0 || cc > 127 )
			return (-1);
		typ = MMCMDTYP_CCV + cc;
	} else
	{
		static SEA_NO tbl[] =
		{
			{ _caf_     ,	MMCMDTYP_CAF			},
			{ _cafdelay_,	MMCMDTYP_CAFDELAY		},
			{ _cafpres_ ,	MMCMDTYP_CAFPRES		},
			{ _cc_      ,	MMCMDTYP_CC				},
			{ "cd"      ,	MMCMDTYP_CAFDELAY		},
			{ _ch_      ,	MMCMDTYP_CH				},
			{ _clock_   ,	MMCMDTYP_CLOCK			},
			{ "d"       ,	MMCMDTYP_DELAY			},
			{ _delay_   ,	MMCMDTYP_DELAY			},
			{ "g"       ,	MMCMDTYP_GATETIME		},
			{ _gate_    ,	MMCMDTYP_GATETIME		},
			{ _gateper_ ,	MMCMDTYP_GATERATIO		},
			{ "gp"      ,	MMCMDTYP_GATERATIO		},
			{ "k"       ,	MMCMDTYP_KEYSHIFT		},
			{ _keyshift_,	MMCMDTYP_KEYSHIFT		},
			{ "o"       ,	MMCMDTYP_OCTAVE			},
			{ "octave"  ,	MMCMDTYP_OCTAVE			},
			{ _offvel_  ,	MMCMDTYP_OFFVELOCITY	},
			{ "ov"      ,	MMCMDTYP_OFFVELOCITY	},
			{ "p"       ,	MMCMDTYP_PITCH			},
			{ _paf_     ,	MMCMDTYP_PAF			},
			{ _pafdelay_,	MMCMDTYP_PAFDELAY		},
			{ _pafpres_ ,	MMCMDTYP_PAFPRES		},
			{ "pd"      ,	MMCMDTYP_PAFDELAY		},
			{ "pg"      ,	MMCMDTYP_PROG			},
			{ _pitch_   ,	MMCMDTYP_PITCH			},
			{ _prog_    ,	MMCMDTYP_PROG			},
			{ "s"       ,	MMCMDTYP_STEPTIME		},
			{ _step_    ,	MMCMDTYP_STEPTIME		},
			{ "t"       ,	MMCMDTYP_TEMPO			},
			{ _tempo_   ,	MMCMDTYP_TEMPO			},
			{ "v"      ,	MMCMDTYP_VELOCITY		},
			{ _vel_     ,	MMCMDTYP_VELOCITY		},
		};
		SEA_NO			key, *keyptr;

		key.name = *s;
		if ( (keyptr = bsearch( &key, tbl, sizeof(tbl)/sizeof(*tbl),
		  sizeof(*tbl), (int (*)(const void *, const void *))compfunc_get_mmlcmd)) != NULL )
		{
			typ = keyptr->no;
			(*s) += strlen(keyptr->name);
		}
	}
	return (typ);
}

static int		compfunc_get_mmlcmd( const SEA_NO *key1, const SEA_NO *key2)
{
	return (mm_strcmpi(key1->name, key2->name));
}

/*************************************************************************
*	ȯ
*************************************************************************/

static long	jrand;
static UINT	ia[56];  /* ia[1..55] */

#define	irn55()	{				\
	int		i;					\
	UINT	j;					\
	for ( i = 1; i <= 24; i++)	\
	{	j = ia[i] - ia[i + 31];	\
		ia[i] = j;				\
	}							\
	for ( i = 25; i <= 55; i++)	\
	{	j = ia[i] - ia[i - 24];	\
		ia[i] = j;				\
	}							\
}

void _FAST	mm_srand( UINT seed )
{
	int		i, ii;
	UINT	k;

	ia[55] = seed;
	k = 1;
	for ( i = 1; i <= 54; i++ )
	{
		ii = (21 * i) % 55;
		ia[ii] = k;
		k = seed - k;
		seed = ia[ii];
	}
	irn55();  irn55();  irn55();  /* warm up */
	jrand = 55;
}

UINT _FAST	mm_rand(void)
{
	if ( ++jrand > 55 )
	{
		irn55();
		jrand = 1;
	}
	return ia[ jrand ];
}
