/*************************************************************************
*	he386p
*************************************************************************/

#include	<stdio.h>
#include	<stdlib.h>
#include	<ctype.h>
#include	<string.h>

#include	"mmoption.h"
#include	"he386.h"
#include	"mmfile.h"
#include	"mml.h"
#include	"mmlib.h"
#include	"mmsg.h"
#include	"msgdat.h"
#include	"cci.h"

#if	defined(__HE386__)
#	include	<spgetopt.h>
#endif
#if	defined(__HE32__)
#	include	"getopt32.h"
#endif

#if	defined(__HE386__) || defined(__HE32__)
	size_t		CodeSize = 128 * 1024;
	size_t		HeepSize =  64 * 1024;
#else
	size_t		CodeSize = 16 * 1024;
	size_t		HeepSize = 16 * 1024;
#endif

LISTWORK		ListMac;

char			CciPath[256]     = {'\0'};
char			CciFileName[256] = "he386.cci";
int				TypMsg = _MSG_ENG;

/*************************************************************************
*	ץѥ᡼βϡ
*************************************************************************/

static void		paraErr( const char *s );
static void		help(void);

int		set_option( int argc, char **argv )
{
	int			n;

	if ( argc < 2 )
		return (0);

	MMsqr.output.cns = 1;
	listwork_init( &ListMac );

	for ( n = 1; n < argc; ++n )
	{
		const char		   *s;

		s = argv[n];
		if ( *s == '-' || *s == '/' )
		{	/* ץ	*/
			++s;	/* skip option switch */
			if ( *s == '?' || strcmpi(s, "help") == 0 )
			{
				help();
				goto _ERR;
			}
			if ( strcmpi(s,"scnMsg") == 0 )
			{
				MMsqr.output.src = 1;
			} else if ( strcmpi(s,"mppMsg") == 0 )
			{
				MMsqr.output.mpp = 1;
			} else if ( strcmpi(s,"dumpMsg") == 0 )
			{
				MMsqr.output.dmp = 1;
			} else if ( strcmpi(s,"debugMsg") == 0 )
			{
				MMsqr.output.deb = 1;
			} else if ( strcmpi(s,"remarkMsg") == 0 )
			{
				MMsqr.output.rem = 1;
			} else if ( strcmpi(s,"cnsMsgIgn") == 0 )
			{
				MMsqr.output.cns = 0;
			} else if ( strcmpi(s,"errOut") == 0 )
			{
				s = argv[n+1];
				if ( n + 1 < argc && *s != '-' && *s != '/' )
				{
					++n;
					strncpy( MMsqr.errFn, s, sizeof(MMsqr.errFn));
				}
			} else if ( strcmpi(s,"msgOut") == 0 )
			{
				s = argv[n+1];
				if ( n + 1 < argc && *s != '-' && *s != '/' )
				{
					++n;
					strncpy( MMsqr.infFn, s, sizeof(MMsqr.errFn));
				}

			} else if ( strcmpi(s,"cciPath") == 0 )
			{
				s = argv[n+1];
				if ( n + 1 < argc && *s != '-' && *s != '/' )
				{
					++n;
					strncpy( CciPath, s, sizeof(CciPath));
				}
			} else if ( strcmpi(s,"cciFile") == 0 )
			{
				s = argv[n+1];
				if ( n + 1 < argc && *s != '-' && *s != '/' )
				{
					++n;
					strncpy( CciFileName, s, sizeof(CciFileName));
				}
			} else if ( strcmpi(s,"codeSize") == 0 )
			{
				s = argv[n+1];
				if ( n + 1 < argc && *s != '-' && *s != '/' )
				{
					unsigned int	v;

					++n;
					v = atoi(s);
					CodeSize = v * 1024;
				}
			} else if ( strcmpi(s,"heepSize") == 0 )
			{
				s = argv[n+1];
				if ( n + 1 < argc && *s != '-' && *s != '/' )
				{
					unsigned int	v;

					++n;
					v = atoi(s);
					HeepSize = v * 1024;
				}

			} else if ( strcmpi(s,"msgTyp") == 0 )
			{
				s = argv[n+1];
				if ( n + 1 < argc && *s != '-' && *s != '/' )
				{
					++n;
					if (      strcmpi(s,"japan"  ) == 0 || strcmpi(s,"jpn") == 0 )
						TypMsg = _MSG_JPN;
					else if ( strcmpi(s,"english") == 0 || strcmpi(s,"eng") == 0 )
						TypMsg = _MSG_ENG;
				}

			} else if ( strcmpi(s,"workMode") == 0 )
			{
				s = argv[n+1];
				if ( n + 1 < argc && *s != '-' && *s != '/' )
				{
					++n;
					if ( strcmpi(s,"memory") == 0 || strcmpi(s,"mem") == 0 )
						MMsqr.trkWork.mode = MMTRKWORK_MODE_MEMORY;
					else if ( strcmpi(s,"file") == 0 )
						MMsqr.trkWork.mode = MMTRKWORK_MODE_FILE;
				}
			} else if ( strcmpi(s,"workPath") == 0 )
			{
				s = argv[n+1];
				if ( n + 1 < argc && *s != '-' && *s != '/' )
				{
					size_t	len;

					++n;
					strncpy( MMsqr.trkWork.path, s, sizeof(MMsqr.trkWork.path));
					len = strlen( MMsqr.trkWork.path );
					if ( len > 0 && !FS_ISDELIM(MMsqr.trkWork.path[len-1]) )
					{
						MMsqr.trkWork.path[len++] = '/';
						MMsqr.trkWork.path[len]   = '\0';
					}
				}

			} else if ( strcmpi(s,"def") == 0 )
			{
				s = argv[n+1];
				if ( n + 1 < argc && *s != '-' && *s != '/' )
				{
					++n;
					list_linkPutf( &ListMac, "%s", s);
				}
			} else
			{
				paraErr( argv[n] );
				goto _ERR;
			}
		} else
		{
			if ( MMsqr.inpFn[0] == '\0' )
				strcpy( MMsqr.inpFn, s );
			else if ( MMsqr.outFn[0] == '\0' )
				strcpy( MMsqr.outFn, s );
		}
	}

	if ( MMsqr.inpFn[0] == '\0' )
	{
		paraErr(NULL);
		goto _ERR;
	}
	return (0);

_ERR:
	return (-1);
}

static void		paraErr( const char *s )
{
	USR_printf("@ Parameter error!!");
	if ( s != NULL )
		USR_printf(" (%s)", s);
	USR_putc('\n');
}

static void		help(void)
{
	USR_printf("\n***** HE386 ver." _MAIN_VERSION " " _MAIN_EDITION " : MML compiler *****\n\n");

#if	defined(__HE32__)
	USR_printf("usage : he386 [<opt>] <file[.mml]> [<file[.mid]>]\n");
#else
	USR_printf("usage : he386p [<opt>] <file[.mml]> [<file[.mid]>]\n");
#endif
	USR_printf("    <opt> : -scnMsg             Output screen message\n");
	USR_printf("            -mppMsg             Output pre-processer message\n");
	USR_printf("            -dumpMsg            Output dump message\n");
	USR_printf("            -cnsMsgIgn          Ignore console message\n");
	USR_printf("            -errOut <file>      Error message output file\n");
	USR_printf("            -MsgOut <file>      Message output file\n");
	USR_printf("            -codeSize <n>       CCI code size (K bytes)\n");
	USR_printf("            -heapSize <n>       CCI heap size (K bytes)\n");
/*	USR_printf("            -cciPath <file>     CCI program pathname\n");	*/
	USR_printf("            -cciFile <file>     CCI program filename\n");
	USR_printf("            -def <name>[=<str>] Macro define\n");
}
