/*************************************************************************
*	ץॢ
*************************************************************************/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>

#include	"he386.h"
#include	"midicc.h"
#include	"mml.h"
#include	"mmac.h"
#include	"mmlib.h"
#include	"mmpart.h"
#include	"mmtrk.h"
#include	"msgdat.h"

/*	config	------------------------------------------------------------*/
/*	#define	_DEBUG	*/
/*--------------------------------------------------------------------*/

static int		progasn_set_vel(MMTRK_T *trk, int argc, char **argv);
static int		progasn_set_note(MMTRK_T *trk, int argc, char **argv);
static int		progasn_set_pitch(MMTRK_T *trk, int argc, char **argv);
static int		progasn_set_cc(MMTRK_T *trk, int argc, char **argv);

/*************************************************************************
*	ץॢ
*-------------------------------------------------------------------------
*	@PROGASN(<cmd>,<mode>,<par>)
*		<cmd>	vel
*				pitch
*				cc:<cc>
*		<mode>	abs		
*				off		եå
*************************************************************************/

int		MMdec_trkProgAsn(MMSQR_T *sqr, MMPART_T *part, int argc, char **argv )
{
	MMTRK_T		*trk = part->trk;
	static struct
	{
		CONST char	*name;
		int			(*func)(MMTRK_T *,int, char **);
	}  tbl[] =
	{
		{ _vel_    , progasn_set_vel  },
		{ _note_   , progasn_set_note },
		{ _pitch_  , progasn_set_pitch},
		{ _cc_     , progasn_set_cc   },
	};

	if ( argc < 3 )
	{
/*_PROGASN_ERR:*/
		sqr->errMsg = "Program assign error!";
		return (ERR);
	}

	{
		int		i;
		char	*cmdname = argv[0];

		for ( i = 0; i < sizeof(tbl)/sizeof(*tbl); ++i )
		{
			int		len;

			len = strlen(tbl[i].name);
			if ( mm_strncmpi( cmdname, tbl[i].name, len) == 0 )
			{
				if ( (tbl[i].func)(trk,argc,argv) )
					return (ERR);
				break;
			}
		}
	}

	return (NORMAL);
}

static int		progasn_set_vel(MMTRK_T *trk, int argc, char **argv)
{
	MMPROGASN_T	*progAsn = &(trk->progAsn);
	int			i;
	int			mode = (strcmpi(argv[1],_off_) == 0) ? TRUE : FALSE;

	if ( mode )
		progAsn->off.vel_flag = TRUE;
	else
		progAsn->abs.vel_flag = TRUE;

	for ( i = 2; i < argc; ++i )
	{
		int			prog, p1, p2, val;
		CONST char	*s;

		s = argv[i];
		if ( mm_getprog_range(&p1,&p2,&s) )
			return (ERR);
		if ( *s != ':' )
			return (ERR);
		++s;
		val = atoi(s);
		if ( mode )
		{	/* եå	*/
			if ( val < -127 || val > 127 )
				return (ERR);
			if ( progAsn->off.vel == NULL )
			{
				REG int		i;
				if ( (progAsn->off.vel = malloc(sizeof(char) * 128)) == NULL )
				{
					((MMSQR_T *)(trk->sqr))->errMsg = _OUT_OF_MEMORY;
					return (ERR);
				}
				for ( i = 0; i < 128; ++i )
					progAsn->off.vel[i] = 0;
			}
			for ( prog = p1; prog <= p2; ++prog )
				progAsn->off.vel[prog] = val;
		} else
		{	/* 		*/
			if ( val < 1 || val > 127 )
				return (ERR);
			if ( progAsn->abs.vel == NULL )
			{
				REG int		i;
				if ( (progAsn->abs.vel = malloc(sizeof(char) * 128)) == NULL )
				{
					((MMSQR_T *)(trk->sqr))->errMsg = _OUT_OF_MEMORY;
					return (ERR);
				}
				for ( i = 0; i < 128; ++i )
					progAsn->abs.vel[i] = -1;
			}
			for ( prog = p1; prog <= p2; ++prog )
				progAsn->abs.vel[prog] = val;
		}
	}
	return (NORMAL);
}

static int		progasn_set_note(MMTRK_T *trk, int argc, char **argv)
{
	MMPROGASN_T	*progAsn = &(trk->progAsn);
	int			i;
	int			mode = (strcmpi(argv[1],_off_) == 0) ? TRUE : FALSE;

	if ( mode )
		progAsn->off.note_flag = TRUE;
	else
		progAsn->abs.note_flag = TRUE;

	for ( i = 2; i < argc; ++i )
	{
		int			prog, p1, p2, val;
		CONST char	*s;

		s = argv[i];
		if ( mm_getprog_range(&p1,&p2,&s) )
			return (ERR);
		if ( *s != ':' )
			return (ERR);
		++s;
		val = atoi(s);
		if ( mode )
		{	/* եå	*/
			if ( val < -127 || val > 127 )
				return (ERR);
			if ( progAsn->off.note == NULL )
			{
				REG int		i;
				if ( (progAsn->off.note = malloc(sizeof(char) * 128)) == NULL )
				{
					((MMSQR_T *)(trk->sqr))->errMsg = _OUT_OF_MEMORY;
					return (ERR);
				}
				for ( i = 0; i < 128; ++i )
					progAsn->off.note[i] = 0;
			}
			for ( prog = p1; prog <= p2; ++prog )
				progAsn->off.note[prog] = val;
		} else
		{	/* 		*/
			if ( val < 1 || val > 127 )
				return (ERR);
			if ( progAsn->abs.note == NULL )
			{
				REG int		i;
				if ( (progAsn->abs.note = malloc(sizeof(char) * 128)) == NULL )
				{
					((MMSQR_T *)(trk->sqr))->errMsg = _OUT_OF_MEMORY;
					return (ERR);
				}
				for ( i = 0; i < 128; ++i )
					progAsn->abs.note[i] = -1;
			}
			for ( prog = p1; prog <= p2; ++prog )
				progAsn->abs.note[prog] = val;
		}
	}
	return (NORMAL);
}

static int		progasn_set_pitch(MMTRK_T *trk, int argc, char **argv)
{
	MMPROGASN_T	*progAsn = &(trk->progAsn);
	int			i;
	int			mode = (strcmpi(argv[1],_off_) == 0) ? TRUE : FALSE;

	if ( mode )
		progAsn->off.pitch_flag = TRUE;
	else
		progAsn->abs.pitch_flag = TRUE;

	for ( i = 2; i < argc; ++i )
	{
		int			prog, p1, p2, val;
		CONST char	*s;

		s = argv[i];
		if ( mm_getprog_range(&p1,&p2,&s) )
			return (ERR);
		if ( *s != ':' )
			return (ERR);
		++s;
		val = atoi(s);
		if ( mode )
		{	/* եå	*/
			if ( val < -16383 || val > 16383 )
				return (ERR);
			if ( progAsn->off.pitch == NULL )
			{
				REG int		i;
				if ( (progAsn->off.pitch = malloc(sizeof(SHORT) * 16384)) == NULL )
				{
					((MMSQR_T *)(trk->sqr))->errMsg = _OUT_OF_MEMORY;
					return (ERR);
				}
				for ( i = 0; i < 16384; ++i )
					progAsn->off.pitch[i] = 0;
			}
			for ( prog = p1; prog <= p2; ++prog )
				progAsn->off.pitch[prog] = val;
		} else
		{	/* 		*/
			if ( val < -8192 || val > 8191 )
				return (ERR);
			if ( progAsn->abs.pitch == NULL )
			{
				REG int		i;
				if ( (progAsn->abs.pitch = malloc(sizeof(SHORT) * 16384)) == NULL )
				{
					((MMSQR_T *)(trk->sqr))->errMsg = _OUT_OF_MEMORY;
					return (ERR);
				}
				for ( i = 0; i < 16384; ++i )
					progAsn->abs.pitch[i] = -1;
			}
			for ( prog = p1; prog <= p2; ++prog )
				progAsn->abs.pitch[prog] = val + 8192;
		}
	}
	return (NORMAL);
}

static int		progasn_set_cc(MMTRK_T *trk, int argc, char **argv)
{
	MMPROGASN_T	*progAsn = &(trk->progAsn);
	int			i, cc, base;
	int			mode = (strcmpi(argv[1],_off_) == 0) ? TRUE : FALSE;

	{
		CONST char	*s = argv[0] ;

		s += 2;	/* skip 'CC' */
		if ( *s != ':' )
			return (ERR);
		++s;
		if ( mm_getIntDigit( &cc, &s) || cc < 0 || cc > 127 )
			return (ERR);
		base = trk->tbl.ccvBase[cc];
	}

	if ( mode )
		progAsn->off.cc[cc].cc_flag = TRUE;
	else
		progAsn->abs.cc[cc].cc_flag = TRUE;

	for ( i = 2; i < argc; ++i )
	{
		int			prog, p1, p2, val;
		CONST char	*s;

		s = argv[i];
		if ( mm_getprog_range(&p1,&p2,&s) )
			return (ERR);
		if ( *s != ':' )
			return (ERR);
		++s;
		val = atoi(s);
		if ( mode )
		{	/* եå	*/
			if ( val < -127 || val > 127 )
				return (ERR);
			if ( progAsn->off.cc[cc].ccv == NULL )
			{
				REG int		i;
				if ( (progAsn->off.cc[cc].ccv = malloc(sizeof(char) * 128)) == NULL )
				{
					((MMSQR_T *)(trk->sqr))->errMsg = _OUT_OF_MEMORY;
					return (ERR);
				}
				for ( i = 0; i < 128; ++i )
					progAsn->off.cc[cc].ccv[i] = 0;
			}
			for ( prog = p1; prog <= p2; ++prog )
				progAsn->off.cc[cc].ccv[prog] = val;
		} else
		{	/* 		*/
			if ( val < -base || val > 127 - base )
				return (ERR);
			if ( progAsn->abs.cc[cc].ccv == NULL )
			{
				REG int		i;
				if ( (progAsn->abs.cc[cc].ccv = malloc(sizeof(char) * 128)) == NULL )
				{
					((MMSQR_T *)(trk->sqr))->errMsg = _OUT_OF_MEMORY;
					return (ERR);
				}
				for ( i = 0; i < 128; ++i )
					progAsn->abs.cc[cc].ccv[i] = -1;
			}
			for ( prog = p1; prog <= p2; ++prog )
				progAsn->abs.cc[cc].ccv[prog] = val + base;	/* Change 1996/02/09 */
		}
	}
	return (NORMAL);
}
